import numpy as np
from .. import math

# TODO: Add smoothing to the min/max/abs/piecewise functions

# TODO: Create test scripts to plot how the different loss model functions behave when smoothing is applied / verify that the behavior is reasonable beyond their intended range
# TODO: Raise warninings to a log file to be aware when the loss model is outside its range of validity and because of which variables
# TODO: the log object can be passed as an additional variable to the functions with a default parameter Logger=None
# TODO: Whenever something is out of range we can do "if Logger: " to check if the logger was actually passed as an input and log an descriptive message

# TODO: Write detailed docstring in .rst format explaning the equations of the loss model?
# TODO: This could be prepared in latex and then transformed into .rst with chatGPT

# TODO: For all functions:
# TODO: add docstring explaning the equations and the original paper
# TODO: explain smoothing/blending tricks
# We can sit together one day and prepare a draft of the explanation of each function with help of chatGPT
# In this way I also get the chance to understand how the loss model works in detail


def compute_losses(input_parameters):
    r"""

    Evaluate loss coefficient according to :cite:`benner_influence_1997`, :cite:`benner_empirical_2006-1` and :cite:`benner_empirical_2006`.

    This model split the total loss coefficient into profile, secondary, trailing edge, tip clearance and incidence losses.
    The loss coefficient are combined through the following relation:

    .. math::

        \mathrm{Y_{tot}} = (\mathrm{Y_{p}} + \mathrm{Y_{te}} + \mathrm{Y_{inc}})(1-Z/H) + \mathrm{Y_{s}} + \mathrm{Y_{cl}}

    where :math:`Z` is the penetration depth of the passage vortex separation line in spanwise direction, and :math:`H` is the mean blade height. This quantity depend on the displacement thickness of the
    inlet endwall bounary layer, which is approximated from a reference value:

    .. math::

        \frac{\delta}{H} = \frac{\delta}{H}_\mathrm{ref} \frac{\mathrm{Re_{in}}}{3\cdot10^5}^{-1/7}

    The reference value may be specified by the user.

    The function calls a function for each loss component:

        - :math:`\mathrm{Y_{p}}` : `get_profile_loss`
        - :math:`\mathrm{Y_{te}}` : `get_trailing_edge_loss`
        - :math:`\mathrm{Y_{s}}` : `get_secondary_loss`
        - :math:`\mathrm{Y_{cl}}` : `get_tip_clearance_loss`
        - :math:`\mathrm{Y_{inc}}` : `get_incidence_loss`
        - :math:`Z/H` : `get_penetration_depth`


    Parameters
    ----------
    input_parameters : dict
        A dictionary containing containing the keys, `flow`, `geometry` and `options`.

    Returns
    -------
    dict
        A dictionary with the loss components.

    """

    # Extract input parameters
    flow_parameters = input_parameters["flow"]
    geometry = input_parameters["geometry"]
    options = input_parameters["loss_model"]

    # Assume angle of minimum incidence loss is equal to metal angle
    # TODO: Digitize Figure 2 from :cite:`moustapha_improved_1990`?
    beta_des = geometry["leading_edge_angle"]

    # Calculate inlet displacement thickness to height ratio
    delta_height = options["inlet_displacement_thickness_height_ratio"]
    delta_height = delta_height * (flow_parameters["Re_in"] / 3e5) ** (-1 / 7)

    # Profile loss coefficient
    Y_p = get_profile_loss(flow_parameters, geometry)

    # Trailing edge coefficient
    Y_te = get_trailing_edge_loss(flow_parameters, geometry)

    # Secondary loss coefficient
    Y_s = get_secondary_loss(flow_parameters, geometry, delta_height)

    # Tip clearance loss coefficient
    Y_cl = get_tip_clearance_loss(flow_parameters, geometry)

    # Incidence loss coefficient
    Y_inc = get_incidence_loss(flow_parameters, geometry, beta_des)

    # Penetration depth to blade height ratio
    ZTE = get_penetration_depth(flow_parameters, geometry, delta_height)

    # Correct profile losses according to penetration depth
    Y_p *= 1 - ZTE
    Y_te *= 1 - ZTE
    Y_inc *= 1 - ZTE

    # Return a dictionary of loss components
    losses = {
        "loss_profile": Y_p,
        "loss_incidence": Y_inc,
        "loss_trailing": Y_te,
        "loss_secondary": Y_s,
        "loss_clearance": Y_cl,
        "loss_total": Y_p + Y_te + Y_inc + Y_s + Y_cl,
    }

    return losses


def get_profile_loss(flow_parameters, geometry):
    r"""
    Calculate the profile loss coefficient for the current cascade using the Kacker and Okapuu loss model.
    The equation for :math:`\mathrm{Y_p}` is given by:

    .. math::

        \mathrm{Y_p} = \mathrm{Y_{reaction}} - \left|\frac{\theta_\mathrm{in}}{\beta_\mathrm{out}}\right| \cdot
        \left(\frac{\theta_\mathrm{in}}{\beta_\mathrm{out}}\right) \cdot (\mathrm{Y_{impulse}} - \mathrm{Y_{reaction}})

    where:

        - :math:`\mathrm{Y_{reaction}}` is the reaction loss coefficient computed using Aungier correlation.
        - :math:`\mathrm{Y_{impulse}}` is the impulse loss coefficient computed using Aungier correlation.
        - :math:`\theta_\mathrm{in}` is the inlet metal angle.
        - :math:`\beta_\mathrm{out}` is the exit flow angle.

    The function also applies various corrections based on flow parameters and geometry factors.

    Parameters
    ----------
    flow_parameters : dict
        Dictionary containing flow-related parameters.
    geometry : dict
        Dictionary with geometric parameters.

    Returns
    -------
    float
        Profile loss coefficient.

    """

    # TODO: explain smoothing/blending tricks

    # Load data
    Re = flow_parameters["Re_out"]
    Ma_rel_out = flow_parameters["Ma_rel_out"]
    Ma_rel_in = flow_parameters["Ma_rel_in"]
    p0rel_in = flow_parameters["p0_rel_in"]
    p_in = flow_parameters["p_in"]
    p0rel_out = flow_parameters["p0_rel_out"]
    p_out = flow_parameters["p_out"]
    beta_out = flow_parameters["beta_out"]

    r_ht_in = geometry["hub_tip_ratio_in"]
    s = geometry["pitch"]
    c = geometry["chord"]
    theta_in = geometry["leading_edge_angle"]
    t_max = geometry["maximum_thickness"]
    cascade_type = geometry["cascade_type"]

    # Reynolds number correction factor
    f_Re = (
        (Re / 2e5) ** (-0.4) * (Re < 2e5)
        + 1 * (Re >= 2e5 and Re <= 1e6)
        + (Re / 1e6) ** (-0.2) * (Re > 1e6)
    )

    # Mach number correction factor
    f_Ma = 1 + 60 * (Ma_rel_out - 1) ** 2 * (
        Ma_rel_out > 1
    )  ## TODO smoothing / mach crit

    # Compute losses related to shock effects at the inlet of the cascade
    f_hub = get_hub_to_mean_mach_ratio(r_ht_in, cascade_type)
    a = max(0, f_hub * Ma_rel_in - 0.4)  # TODO: smoothing
    Y_shock = 0.75 * a**1.75 * r_ht_in * (p0rel_in - p_in) / (p0rel_out - p_out)
    Y_shock = max(0, Y_shock)  # TODO: smoothing

    # Compute compressible flow correction factors
    Kp, K2, K1 = get_compressible_correction_factors(Ma_rel_in, Ma_rel_out)

    # Yp_reaction and Yp_impulse according to Aungier correlation
    # These formulas are valid for 40<abs(angle_out)<80
    # Extrapolating outside of this limits might give completely wrong results
    # If the optimization algorithm has upper and lower bounds for the outlet
    # angle there is no need to worry about this problem
    # angle_out_bis keeps the 40deg-losses for outlet angles lower than 40deg
    angle_out_bis = max(abs(beta_out), 40)
    Yp_reaction = nozzle_blades(s / c, angle_out_bis)
    Yp_impulse = impulse_blades(s / c, angle_out_bis)

    # Formula according to Kacker-Okapuu
    Y_p = Yp_reaction - abs(theta_in / beta_out) * (theta_in / beta_out) * (
        Yp_impulse - Yp_reaction
    )

    # Limit the extrapolation of the profile loss to avoid negative values for
    # blade profiles with little deflection
    # Low limit to 80% of the axial entry nozzle profile loss
    # This value is completely arbitrary
    Y_p = max(Y_p, 0.8 * Yp_reaction)  # TODO: smoothing

    # Avoid unphysical effect on the thickness by defining the variable aa
    aa = max(0, -theta_in / beta_out)  # TODO: smoothing
    Y_p = Y_p * ((t_max / c) / 0.2) ** aa
    Y_p = 0.914 * (2 / 3 * Y_p * Kp + Y_shock)

    # Corrected profile loss coefficient
    Y_p = f_Re * f_Ma * Y_p

    return Y_p


def get_trailing_edge_loss(flow_parameters, geometry):
    r"""
    Calculate the trailing edge loss coefficient using the Kacker-Okapuu model.
    The main equation for the kinetic-energy coefficient is given by:

    .. math::

        d_{\phi^2} = d_{\phi^2_\mathrm{reaction}} - \left|\frac{\theta_\mathrm{in}}{\beta_\mathrm{out}}\right| \cdot
        \left(\frac{\theta_\mathrm{in}}{\beta_\mathrm{out}}\right) \cdot (d_{\phi^2_\mathrm{impulse}} - d_{\phi^2_\mathrm{reaction}})

    The kinetic-energy coefficient is converted to the total pressure loss coefficient by:

    .. math::

        \mathrm{Y_{te}} = \frac{1}{{1 - \phi^2}} - 1

    where:

        - :math:`d_{\phi^2_\mathrm{reaction}}` and :math:`d_{\phi^2_\mathrm{impulse}}` are coefficients related to kinetic energy loss for reaction and impulse blades respectively, and are interpolated based on trailing edge to throat opening ratio.
        - :math:`\beta_\mathrm{out}` is the exit flow angle.
        - :math:`\theta_\mathrm{in}` is the inlet metal angle.

    The function also applies various interpolations and computations based on flow parameters and geometry factors.

    Parameters
    ----------
    flow_parameters : dict
        Dictionary containing flow-related parameters.
    geometry : dict
        Dictionary with geometric parameters.

    Returns
    -------
    float
        Trailing edge loss coefficient.

    """

    t_te = geometry["trailing_edge_thickness"]
    o = geometry["opening"]
    angle_in = geometry["leading_edge_angle"]
    angle_out = flow_parameters["beta_out"]

    # Range of trailing edge to throat opening ratio
    r_to_data = [0, 0.2, 0.4]

    # Reacting blading
    phi_data_reaction = [0, 0.045, 0.15]

    # Impulse blading
    phi_data_impulse = [0, 0.025, 0.075]

    # Numerical trick to avoid too big r_to's
    r_to = min(0.4, t_te / o)  # TODO: smoothing

    # Interpolate data
    d_phi2_reaction = np.interp(r_to, r_to_data, phi_data_reaction)
    d_phi2_impulse = np.interp(r_to, r_to_data, phi_data_impulse)

    # Compute kinetic energy loss coefficient
    d_phi2 = d_phi2_reaction - abs(angle_in / angle_out) * (angle_in / angle_out) * (
        d_phi2_impulse - d_phi2_reaction
    )

    # Limit the extrapolation of the trailing edge loss
    d_phi2 = max(d_phi2, d_phi2_impulse / 2)  # TODO: smoothing
    Y_te = 1 / (1 - d_phi2) - 1

    return Y_te


def get_secondary_loss(flow_parameters, geometry, delta_height):
    r"""
    Calculate the secondary loss coefficient.

    The correlation takes two different forms depending on the aspect ratio:

    If aspect ratio :math:`\leq 2.0`

    .. math::

        \mathrm{Y_{s}} = \frac{0.038 + 0.41 \tanh(1.20\delta^*/H)}{\sqrt{\cos(\xi)}CR(H/c)^{0.55}\frac{\cos(\beta_\mathrm{out})}{\cos(\xi)}^{0.55}}

    wheras if aspect ratio :math:`>2.0`

    .. math::

        \mathrm{Y_{s}} = \frac{0.052 + 0.56 \tanh(1.20\delta^*/H)}{\sqrt{\cos(\xi)}CR(H/c)\frac{\cos(\beta_\mathrm{out})}{\cos(\xi)}^{0.55}}

    where:

        - :math:`CR = \frac{\cos(\beta_\mathrm{in})}{\cos(\beta_\mathrm{out})}` is the convergence ratio.
        - :math:`H` is the mean height.
        - :math:`c` is the chord.
        - :math:`\delta^*` is the inlet endwall boundary layer displacement thickness.
        - :math:`\xi` is the stagger angle.
        - :math:`\beta_\mathrm{out}` is the exit relative flow angle.

    Parameters
    ----------
    flow_parameters : dict
        Dictionary containing flow-related parameters.
    geometry : dict
        Dictionary with geometric parameters.

    Returns
    -------
    float
        Secondary loss coefficient.

    """

    beta_in = flow_parameters["beta_in"]
    beta_out = flow_parameters["beta_out"]
    height = geometry["height"]
    chord = geometry["chord"]
    stagger = geometry["stagger_angle"]

    AR = height / chord
    CR = math.cosd(beta_in) / math.cosd(beta_out)
    if AR <= 2:  # TODO: sigmoid blending to convert to smooth piecewise function
        denom = (
            np.sqrt(math.cosd(stagger))
            * CR
            * AR**0.55
            * (math.cosd(beta_out) / (math.cosd(stagger))) ** 0.55
        )
        Y_sec = (0.038 + 0.41 * np.tanh(1.2 * delta_height)) / denom
    else:
        denom = (
            np.sqrt(math.cosd(stagger))
            * CR
            * AR
            * (math.cosd(beta_out) / (math.cosd(stagger))) ** 0.55
        )
        Y_sec = (0.052 + 0.56 * np.tanh(1.2 * delta_height)) / denom

    return Y_sec


def get_tip_clearance_loss(flow_parameters, geometry):
    r"""
    Calculate the tip clearance loss coefficient for the current cascade using the Kacker and Okapuu loss model.
    The equation for the tip clearance loss coefficent is given by:

    .. math::

        \mathrm{Y_{cl}} = B \cdot Z \cdot \frac{c}{H} \cdot \left(\frac{t_\mathrm{cl}}{H}\right)^{0.78}

    where:

        - :math:`B` is an empirical parameter that depends on the type of cascade (0 for stator, 0.37 for shrouded rotor).
        - :math:`Z` is a blade loading parameter
        - :math:`c` is the chord.
        - :math:`H` is the mean blade height.
        - :math:`t_\mathrm{cl}` is the tip clearance.


    Parameters
    ----------
    flow_parameters : dict
        Dictionary containing flow-related parameters.
    geometry : dict
        Dictionary with geometric parameters.

    Returns
    -------
    float
        Tip clearance loss coefficient.
    """

    beta_out = flow_parameters["beta_out"]
    beta_in = flow_parameters["beta_in"]

    H = geometry["height"]
    c = geometry["chord"]
    t_cl = geometry["tip_clearance"]
    cascade_type = geometry["cascade_type"]

    # Calculate blade loading parameter Z
    angle_m = math.arctand((math.tand(beta_in) + math.tand(beta_out)) / 2)
    Z = (
        4
        * (math.tand(beta_in) - math.tand(beta_out)) ** 2
        * math.cosd(beta_out) ** 2
        / math.cosd(angle_m)
    )

    # Empirical parameter (0 for stator, 0.37 for shrouded rotor)
    if cascade_type == "stator":
        B = 0
    elif cascade_type == "rotor":
        B = 0.37
    else:
        print("Specify the type of cascade")

    # Tip clearance loss coefficient
    Y_cl = B * Z * c / H * (t_cl / H) ** 0.78

    return Y_cl


def get_incidence_loss(flow_parameters, geometry, beta_des):
    r"""
    Calculate the incidence loss coefficient according to the correlation proposed by :cite:`benner_influence_1997`.

    This model calculates the incidence loss parameter through the function `get_incidence_parameter`. The kinetic energu coefficient
    can be obtained from `get_incidence_profile_loss_increment`, which is a function of the incidence parameter. The kinetic energy loss coefficient is converted
    to the total pressure loss coefficient through `convert_kinetic_energy_coefficient`.

    Parameters
    ----------
    flow_parameters : dict
        Dictionary containing flow-related parameters.
    geometry : dict
        Dictionary with geometric parameters.
    beta_des : float
        Inlet flow angle for zero incidence losses (design).

    Returns
    -------
    float
        Incidence loss coefficient.
    """

    beta_in = flow_parameters["beta_in"]
    gamma = flow_parameters["gamma_out"]
    Ma_rel_out = flow_parameters["Ma_rel_out"]

    le = geometry["leading_edge_diameter"]
    s = geometry["pitch"]
    We = geometry["leading_edge_wedge_angle"]
    theta_in = geometry["leading_edge_angle"]
    theta_out = math.arccosd(geometry["A_throat"] / geometry["A_out"])

    chi = get_incidence_parameter(le, s, We, theta_in, theta_out, beta_in, beta_des)

    dPhip = get_incidence_profile_loss_increment(chi)

    Y_inc = convert_kinetic_energy_coefficient(dPhip, gamma, Ma_rel_out)

    return Y_inc


def get_hub_to_mean_mach_ratio(r_ht, cascade_type):
    r"""
    Compute the ratio between Mach at the hub and mean span at the inlet of the current cascade.

    Due to radial variation in gas conditions, Mach at the hub will always be higher than at the mean.
    Thus, shock losses at the hub could occur even when the Mach is subsonic at the mean blade span.

    Parameters
    ----------
    r_ht : float
        Hub to tip ratio at the inlet of the current cascade.
    cascade_type : str
        Type of the current cascade, either 'stator' or 'rotor'.

    Returns
    -------
    float
        Ratio between Mach at the hub and mean span at the inlet of the current cascade.

    """

    if r_ht < 0.5:  # TODO: add smoothing, this is essentially a max(r_ht, 0.5)
        r_ht = 0.5  # Numerical trick to prevent extrapolation

    r_ht_data = [0.5, 0.6, 0.7, 0.8, 0.9, 1.0]

    # Stator curve
    f_data_S = [1.4, 1.18, 1.05, 1.0, 1.0, 1.0]

    # Rotor curve
    f_data_R = [2.15, 1.7, 1.35, 1.12, 1.0, 1.0]

    if cascade_type == "stator":
        f = np.interp(r_ht, r_ht_data, f_data_S)
    elif cascade_type == "rotor":
        f = np.interp(r_ht, r_ht_data, f_data_R)
    else:
        print("Specify the type of cascade")

    return f


def get_compressible_correction_factors(Ma_rel_in, Ma_rel_out):
    r"""

    Compute compressible flow correction factor according to Kacker and Okapuu loss model :cite:`kacker_mean_1982`.

    The correction factors :math:`\mathrm{K_1}`, :math:`\mathrm{K_2}` and :math:`\mathrm{K_p}` was introduced by :cite:`kacker_mean_1982` to correct previous correlation (:cite:`ainley_method_1951`)
    for effect of higher mach number and channel acceleration. The correction factors reduces the losses at higher mach number. Their definition follows: 

    .. math::

        &\mathrm{K_1} = \begin{cases}
                1 & \text{if } \mathrm{Ma_{out}} < 0.2 \\
                1 - 1.25(\mathrm{Ma_{out}} - 0.2) & \text{if } \mathrm{Ma_{out}} \geq 0.2
              \end{cases} \\
        &\mathrm{K_2} = \left(\frac{\mathrm{Ma_{in}}}{\mathrm{Ma_{out}}}\right) \\
        &\mathrm{K_p} = 1 - \mathrm{K_2}(1-\mathrm{K_1})

    where:

        - :math:`\mathrm{Ma_{in}}` is the relative mach number at the cascade inlet.
        -  :math:`\mathrm{Ma_{out}}` is the relative mach number at the cascade exit.
    
    Parameters
    ----------
    Ma_rel_in : float
        Inlet relative mach number.
    Ma_rel_out : float
        Exit relative mach number.

    Returns
    -------
    float
        Correction factor :math:`\mathrm{K_p}`.
    float
        Correction factor :math:`\mathrm{K_2}`.
    float
        Correction factor :math:`\mathrm{K_1}`.
    """

    K1 = 1 * (Ma_rel_out < 0.2) + (1 - 1.25 * (Ma_rel_out - 0.2)) * (
        Ma_rel_out > 0.2 and Ma_rel_out < 1.00
    )  # TODO: this can be converted to a smooth piecewise function (sigmoid blending)
    K2 = (Ma_rel_in / Ma_rel_out) ** 2
    Kp = 1 - K2 * (1 - K1)
    Kp = max(0.1, Kp)  # TODO: smoothing
    return [Kp, K2, K1]


def nozzle_blades(r_sc, angle_out):
    r"""
    Use Aungier correlation to compute the pressure loss coefficient for nozzle blades :cite:`aungier_turbine_2006`.

    This correlation is a formula that reproduces the figures from the Ainley and Mathieson original figures :cite:`ainley_method_1951`,
    and is a function of the pitch-to-chord ratio and exit relative flow angle. 

    The correlation uses the following equations:

    .. math::

        & \beta_\mathrm{tan} = 90 - \beta_\mathrm{ax} \\
        & \left(\frac{s}{c}\right)_\mathrm{min} = \begin{cases}
                                                    0.46 + \frac{\beta_\mathrm{tan}}{77} && \text{if } \beta_\mathrm{tan} < 30 \\
                                                    0.614 + \frac{\beta_\mathrm{tan}}{130} && \text{if } \beta_\mathrm{tan} \geq 30
                                                \end{cases} \\
        & X = \left(\frac{s}{c}\right) - \left(\frac{s}{c}\right)_\mathrm{min} \\
        & A = \begin{cases}
                0.025 + \frac{27 - \beta_\mathrm{tan}}{530} && \text{if } \beta_\mathrm{tan} < 27 \\
                0.025 + \frac{27 - \beta_\mathrm{tan}}{3085} && \text{if } \beta_\mathrm{tan} \geq 27
              \end{cases} \\
        & B = 0.1583 - \frac{\beta_\mathrm{tan}}{1640} \\
        & C = 0.08\left(\frac{\beta_\mathrm{tan}}{30}\right)^2 -1 \\
        & n = 1 + \frac{\beta_\mathrm{tan}}{30} \\
        & \mathrm{Y_{p,reaction}} = \begin{cases}
                A + BX^2 + CX^3 && \text{if } \beta_\mathrm{tan} < 30 \\
                A + B |X|^n && \text{if } \beta_\mathrm{tan} \geq 30
              \end{cases}   

    where:

        - :math:`s` is the pitch
        - :math:`c` is the chord
        - :math:`\beta_\mathrm{tan}` and :math:`\beta_\mathrm{ax}` is the exit relative flow angle with respect to tangential and axial direction. 
        
    Parameters
    ----------
    r_sc : float
        Pitch-to-chord ratio.
    angle_out : float
        Exit relative flow angle (in degrees).

    Returns
    -------
    float
        Pressure loss coefficient for impulse blades.

    """

    phi = 90 - angle_out
    r_sc_min = (0.46 + phi / 77) * (phi < 30) + (0.614 + phi / 130) * (
        phi >= 30
    )  # TODO: sigmoid blending to convert to smooth piecewise function
    X = r_sc - r_sc_min
    A = (0.025 + (27 - phi) / 530) * (phi < 27) + (0.025 + (27 - phi) / 3085) * (
        phi >= 27
    )  # TODO: sigmoid blending to convert to smooth piecewise function
    B = 0.1583 - phi / 1640
    C = 0.08 * ((phi / 30) ** 2 - 1)
    n = 1 + phi / 30
    Yp_reaction = (A + B * X**2 + C * X**3) * (phi < 30) + (A + B * abs(X) ** n) * (
        phi >= 30
    )  # TODO: sigmoid blending to convert to smooth piecewise function
    return Yp_reaction


def impulse_blades(r_sc, angle_out):
    r"""
    Use Aungier correlation to compute the pressure loss coefficient for impulse blades :cite:`aungier_turbine_2006`.

    This correlation is a formula that reproduces the figures from the Ainley and Mathieson original figures :cite:`ainley_method_1951`,
    and is a function of the pitch-to-chord ratio and exit relative flow angle. 

    The correlation uses the following equations:

    .. math::

        & \beta_\mathrm{tan} = 90 - \beta_\mathrm{ax} \\
        & \left(\frac{s}{c}\right)_\mathrm{min} = 0.224 + 1.575\left(\frac{\beta_\mathrm{tan}}{90}\right) - \left(\frac{\beta_\mathrm{tan}}{90}\right)^2 \\
        & X = \left(\frac{s}{c}\right) - \left(\frac{s}{c}\right)_\mathrm{min} \\
        & A = 0.242 - \frac{\beta_\mathrm{tan}}{151} + \left(\frac{\beta_\mathrm{tan}}{127}\right)^2 \\
        & B = \begin{cases}
                0.3 + \frac{30 - \beta_\mathrm{tan}}{50} && \text{if } \beta_\mathrm{tan} < 30 \\
                0.3 + \frac{30 - \beta_\mathrm{tan}}{275} && \text{if } \beta_\mathrm{tan} \geq 30
              \end{cases} \\
        & C = 0.88 - \frac{\beta_\mathrm{tan}}{42.4} + \left(\frac{\beta_\mathrm{tan}}{72.8}\right)^2 \\
        & \mathrm{Y_{p,impulse}} = A + BX^2 - CX^3 

    where:

        - :math:`s` is the pitch
        - :math:`c` is the chord
        - :math:`\beta_\mathrm{tan}` and :math:`\beta_\mathrm{ax}` is the exit relative flow angle with respect to tangential and axial direction. 
        
    Parameters
    ----------
    r_sc : float
        Pitch-to-chord ratio.
    angle_out : float
        Exit relative flow angle (in degrees).

    Returns
    -------
    float
        Pressure loss coefficient for impulse blades.

    """

    phi = 90 - angle_out
    r_sc_min = 0.224 + 1.575 * (phi / 90) - (phi / 90) ** 2
    X = r_sc - r_sc_min
    A = 0.242 - phi / 151 + (phi / 127) ** 2
    B = (0.3 + (30 - phi) / 50) * (phi < 30) + (0.3 + (30 - phi) / 275) * (
        phi >= 30
    )  # TODO: sigmoid blending to convert to smooth piecewise function
    C = 0.88 - phi / 42.4 + (phi / 72.8) ** 2
    Yp_impulse = A + B * X**2 - C * X**3
    return Yp_impulse


def get_penetration_depth(flow_parameters, geometry, delta_height):
    r"""
    Calculated the penetration depth of the passage vortex separation line relative to the blade span.

    The endwall inlet boundary layer generate a vortex which propagtes through the cascade section,
    and the spanwise penetration of this vortex affect the magnutude of the secondary loss coefficient.
    This function approximate the vortex penetration depth to the blade span by the correlation developed by :cite:`benner_empirical_2006-1`.

    The quantity is calculated as:

    .. math::

        \frac{\mathrm{Z_{te}}}{H} = \frac{0.10F_t^{0.79}}{\sqrt{CR}\left(\frac{H}{c}\right)^{0.55}} + 32.70\frac{\delta^*}{H}^2

    where:

        - :math:`CR = \frac{\cos(\beta_\mathrm{in})}{\cos(\beta_\mathrm{out})}` is the convergence ratio.
        - :math:`H` is the mean height.
        - :math:`c` is the chord.
        - :math:`\delta^*` is the inlet endwall boundary layer displacement thickness.
        - :math:`\beta_\mathrm{out}` is the exit relative flow angle.
        - :math:`F_t` is the tangential loading coefficient.

    The tangiential loading coefficient is calculated by a separate function (`F_t`).

    Parameters
    ----------
    flow_parameters : dict
        Dictionary containing flow-related parameters.
    geometry : dict
        Dictionary with geometric parameters.
    delta_height : float
        The inlet endwall boundary layer displacement thickness relative to the mean blade height.

    Returns
    -------
    float
        Spanwise penetration depth of the passage vortex relative to the mean blade height.

    """

    # TODO: explain smoothing/blending tricks

    beta_in = flow_parameters["beta_in"]
    beta_out = flow_parameters["beta_out"]
    axial_chord = geometry["axial_chord"]
    pitch = geometry["pitch"]
    chord = geometry["chord"]
    height = geometry["height"]

    CR = math.cosd(beta_in) / math.cosd(
        beta_out
    )  # Convergence ratio from Benner et al.[2006]

    BSx = axial_chord / pitch  # Axial blade solidity
    delta_height = (
        delta_height  # Boundary layer displacement thickness relative to blade height
    )
    AR = height / chord  # Aspect ratio

    Ft = F_t(BSx, beta_in, beta_out)

    Z_TE = 0.10 * Ft**0.79 / np.sqrt(CR) / (AR) ** 0.55 + 32.70 * delta_height**2

    Z_TE = min(Z_TE, 0.99)  # TODO: smoothing

    return Z_TE


def convert_kinetic_energy_coefficient(dPhi, gamma, Ma_rel_out):
    r"""

    Convert the kinetic energy coefficient increment due to incidence to the total pressure loss coefficient according to the following correlation:

    .. math::

        \mathrm{Y} = \frac{\left(1-\frac{\gamma -1}{2}\mathrm{Ma_{out}}^2(\frac{1}{(1-\Delta\phi^2_p)}-1)\right)^\frac{-\gamma}{\gamma - 1}-1}{1-\left(1 + \frac{\gamma - 1}{2}\mathrm{Ma_{out}}^2\right)^\frac{-\gamma}{\gamma - 1}}

    where:

        - :math:`\gamma` is the specific heat ratio.
        - :math:`\mathrm{Ma_{out}}` is the cascade exit relative mach number.
        - :math:`\Delta\phi^2_p` is the kinetic energy loss coefficient increment due to incidence.

    Parameters
    ----------
    dPhi : float
        Kinetic energy coefficient increment.
    gamma : float
        Heat capacity ratio.
    Ma_rel_out : float
        The cascade exit relative mach number.

    Returns
    -------
    float
        The total pressure loss coefficient.

    Warnings
    --------
    This conversion assumes that the fluid is a perfect gas.

    """

    denom = 1 - (1 + (gamma - 1) / 2 * Ma_rel_out**2) ** (-gamma / (gamma - 1))
    numer = (1 - (gamma - 1) / 2 * Ma_rel_out**2 * (1 / (1 - dPhi) - 1)) ** (
        -gamma / (gamma - 1)
    ) - 1

    Y = numer / denom

    return Y


def get_incidence_profile_loss_increment(chi, chi_extrapolation=5, loss_limit=0.5):
    r"""
    Computes the increment in profile losses due to the effect of incidence according to the correlation proposed by :cite:`benner_influence_1997`.

    The increment in profile losses due to incidence is based on the incidence parameter :math:`\chi`.
    The formula used to compute :math:`\chi` is given by:

    .. math::

       \chi = We^{-0.2}\left(\frac{d}{s}\right)^{-0.05}  \left(\frac{\cos{\beta_1}}{\cos{\beta_2}} \right)^{-1.4}  \left(\alpha_1 - \alpha_{1,\mathrm{des}}\right)

    Depending on the value of :math:`\chi`, two equations are used for computing the increment in profile losses:

    1. For :math:`\chi \geq 0`:

    .. math::

        \Delta \phi_{p}^2 = \sum_{i=1}^{8} a_i \, \chi^i

    with coefficients:
    
    .. math::
    
        a_1 = -6.149 \times 10^{-5}  \\
        a_2 = +1.327 \times 10^{-3}  \\
        a_3 = -2.506 \times 10^{-4}  \\
        a_4 = -1.542 \times 10^{-4}  \\
        a_5 = +9.017 \times 10^{-5}  \\
        a_6 = +1.106 \times 10^{-5}  \\
        a_7 = -5.318 \times 10^{-6}  \\
        a_8 = +3.711 \times 10^{-7}

    2. For :math:`\chi < 0`:

    .. math::

        \Delta \phi_{p}^2 = \sum_{i=1}^{2} b_i \, \chi^i

    with coefficients:
    
    .. math::
    
        b_1 = -8.720e-4 \times 10^{-4}  \\
        b_2 = +1.358e-4 \times 10^{-4}
    
    This function also implements two safeguard methods to prevent unrealistic values of the pressure loss increment
    when the value of :math:`\chi` is outside the range of validity of the correlation

    1. Linear extrapolation beyond a certain :math:`\chi`:
        Beyond a certain value of the distance parameter :math:`\hat{\chi}` the function linearly extrapolates based on the slope of the polynomial.

    2. Loss Limiting:
        The profile loss increment is limited to a maximum value :math:`\Delta \hat{\phi}_{p}^2` to prevent potential over-predictions. The limiting mechanism is smooth to ensure that the function is differentiable.

    The final expression for the loss coefficient increment can be summarized as:

    .. math::

        \begin{align}
        \Delta \phi_{p}^2 = 
        \begin{cases}
        \sum_{i=1}^{2} b_i \, \chi^i & \text{for } \chi < 0 \\
        \sum_{i=1}^{8} a_i \, \chi^i & \text{for } 0 \leq \chi  \leq \hat{\chi} \\
        \Delta\phi_{p}^2(\hat{\chi}) + \mathrm{slope}(\hat{\chi}) \cdot(\chi - \hat{\chi}) & \text{for } \chi > \hat{\chi}
        \end{cases}
        \end{align}

    .. math::
        \Delta \phi_{p}^2 = \mathrm{smooth\,min}(\Delta \phi_{p}^2, \Delta \hat{\phi}_{p}^2)


    Parameters
    ----------
    chi : array-like
        The distance parameter, :math:`\chi`, used to compute the profile losses increment.
    chi_extrapolation : float, optional
        The value of :math:`\chi` beyond which linear extrapolation is applied. Default is 5.
    loss_limit : float, optional
        The upper limit to which the profile loss increment is smoothly constrained. Default is 0.5.

    Returns
    -------
    array-like
        Increment in profile losses due to the effect of incidence.
    """

    # Define the polynomial function for positive values of chi
    coeffs_1 = np.array(
        [
            +3.711e-7,
            -5.318e-6,
            +1.106e-5,
            +9.017e-5,
            -1.542e-4,
            -2.506e-4,
            +1.327e-3,
            -6.149e-5,
        ]
    )
    func_1 = lambda x: np.sum(
        np.asarray([a * x**i for (i, a) in enumerate(coeffs_1[::-1], start=1)]),
        axis=0,
    )

    # Define the polynomial function for negative values of chi
    coeffs_2 = np.array([1.358e-4, -8.72e-4])
    func_2 = lambda x: np.sum(
        np.asarray([b * x**i for (i, b) in enumerate(coeffs_2[::-1], start=1)]),
        axis=0,
    )

    # Calculate the profile losses increment based on polynomials
    loss_poly = np.where(chi >= 0, func_1(chi), func_2(chi))

    # Apply linear extrapolation beyond the threshold for chi
    _chi = chi_extrapolation
    loss = func_1(_chi)
    coeffs_slope = np.array([i * a for i, a in enumerate(coeffs_1[::-1], start=1)])
    slope = np.sum([a * _chi ** (i - 1) for i, a in enumerate(coeffs_slope, start=1)])
    loss_extrap = loss + slope * (chi - _chi)
    loss_increment = np.where(chi > _chi, loss_extrap, loss_poly)

    # Limit the loss to the loss_limit
    if loss_limit is not None:
        loss_increment = np.asarray(
            [loss_increment, np.full_like(loss_increment, loss_limit)]
        )
        loss_increment = math.smooth_min(
            loss_increment, method="logsumexp", alpha=25, axis=0
        )

    return loss_increment


def get_incidence_parameter(le, s, We, theta_in, theta_out, beta_in, beta_des):
    r"""
    Calculate the incidence parameter according to the correlation proposed by :cite:`benner_influence_1997`.

    The incidence parameter is used to calculate the increment in profile losses due to the effect of incidence according to function `get_incidence_profile_loss_increment`.

    The quantity is calculated as:

    .. math::

        \chi = \frac{\mathrm{d_{le}}}{s}^{-0.05}\mathrm{We_{le}}^{-0.2}\frac{\cos\theta_\mathrm{in}}{\cos\theta_\mathrm{out}}^{-1.4}(\beta_\mathrm{in} - \beta_\mathrm{des})

    where:

        - :math:`\mathrm{d_{le}}` is the leading edge diameter.
        - :math:`s` is the pitch.
        - :math:`\mathrm{We_{le}}` is the leading edge wedge angle.
        - :math:`\theta_\mathrm{in}` and :math:`\theta_\mathrm{out}` is the blade metal angle at the inlet and outlet respectively.
        - :math:`\beta_\mathrm{in}` and :math:`\beta_\mathrm{des}` is the inlet relative flow angle at given and design conditions respectively.

    Parameters
    ----------
    le : float
        Leading edge diameter.
    s : float
        Pitch.
    We : float
        Leading edge wedge angle (in degrees).
    theta_in : float
        Leading edge metal angle (in degrees).
    theta_out : float
        Trailing edge metal angle (in degrees).
    beta_in : float
        Inlet relative flow angle (in degrees).
    beta_des : float
        Inlet relative flow angle at design condition (in degrees).

    Returns
    -------
    float
        Incidence parameter.
    """

    chi = (
        (le / s) ** (-0.05)
        * (We) ** (-0.2)
        * (math.cosd(theta_in) / math.cosd(theta_out)) ** (-1.4)
        * (abs(beta_in) - abs(beta_des))
    )
    return chi


def F_t(BSx, beta_in, beta_out):
    r"""

    Calculate the tangential loading coefficient according to the correlation proposed by :cite:`benner_empirical_2006-1`.

    .. math::

        F_t = 2\frac{s}{c_\mathrm{ax}}\cos^2(\beta_m)(\tan(\beta_\mathrm{in} - \beta_\mathrm{out}))

    where:

        - :math:`s` is the pitch.
        - :math:`c_\mathrm{ax}` is the axial chord.
        - :math:`\beta_m = \tan^{-1}(0.5(\tan(\beta_\mathrm{in}) + \tan(\beta_\mathrm{out})))` is the mean vector angle.
        - :math:`\beta_\mathrm{in}` and :math:`\beta_\mathrm{out}` is the inlet and outlet relative flow angle.

    Parameters
    ----------
    Bsx : float
        Blade solidity based on axial chord.
    beta_in : float
        Inlet relative flow angle (in degrees).
    beta_out : float
        Exit relative flow angle (in degrees).

    Returns
    -------
    float
        Tangetnial loading parameter.

    """

    a_m = math.arctand(0.5 * (math.tand(beta_in) + math.tand(beta_out)))

    F_t = (
        2
        * 1
        / BSx
        * math.cosd(a_m) ** 2
        * (abs(math.tand(beta_in)) + abs(math.tand(beta_out)))
    )

    return F_t



import os
import yaml
import copy
import time
import datetime
import itertools
import numpy as np
import pandas as pd
import CoolProp as cp
import matplotlib.pyplot as plt

from .. import math
from .. import pysolver_view as psv
from .. import utilities as utils
from .. import properties as props
from . import geometry_model as geom
from . import flow_model as flow

SOLVER_MAP = {"lm": "Lavenberg-Marquardt", "hybr": "Powell's hybrid"}
"""
Available solvers for performance analysis.
"""


def get_heuristic_guess_input(n):
    r"""

    Generate a list of `n` number of different sets of dictionaries used to generate initial guesses.

    Total-to-total efficiency, total-to-static efficiency, enthalpy loss fractions for each cascade and critical mach can be used to generate
    an initial guess for performance analysis. This function generate a list of such different sets used to generate a full initial guess through the function `compute_heuristic_initial_guess`.

    The total-to-static efficiency, varies between  0.6, 0.7, or 0.8, while the total-to-total efficiency can be 0.7, 0.8 or 0.9. The enthalpy loss fractions can either split equally between the cascade,
    or with equal increments for neighbouring values (see `utils.fill_array_with_increment` for more information). The critical mach is assumed to be 0.95.

    Parameters
     ----------
     n : int
         Number of cascades.

     Returns
     -------
     list
         List of sets of variables used to generate an initial guess.

    """

    eta_ts_vec = [0.8, 0.7, 0.6]

    array = utils.fill_array_with_increment(n)
    enthalpy_distributions = []
    enthalpy_distributions.append(np.ones(n) * 1 / n)
    enthalpy_distributions.append(array)
    enthalpy_distributions.append(np.flip(array))

    initial_guesses = []
    for eta_ts in eta_ts_vec:
        for enthalpy_distribution in enthalpy_distributions:
            initial_guesses.append(
                {
                    "enthalpy_loss_fractions": enthalpy_distribution,
                    "eta_ts": eta_ts,
                    "eta_tt": eta_ts + 0.1,
                    "Ma_crit": 0.95,
                }
            )

    return initial_guesses


def compute_performance(
    operation_points,
    config,
    initial_guess=None,
    out_filename=None,
    out_dir="output",
    stop_on_failure=False,
    export_results=True,
):
    r"""
    Compute and export the performance of each specified operation point to an Excel file.

    This function handles two types of input for operation points:

        1. An explicit list of dictionaries, each detailing a specific operation point.
        2. A dictionary where each key has a range of values, representing the cross-product of all possible operation points. It generates the Cartesian product of these ranges internally.

    For each operation point, it computes performance based on the provided case data and compiles
    the results into an Excel workbook with multiple sheets for various data sections.

    The function validates the input operation points, and if they are given as ranges, it generates
    all possible combinations. Performance is computed for each operation point, and the results are
    then stored in a structured Excel file with separate sheets for each aspect of the data (e.g.,
    overall, plane, cascade, stage, solver, and solution data).

    The initial guess variable is used for the first operation point. If given, it must be a dictionary with the following keys:

        - `enthalpy_loss_fractions`, which is a list containing the assumed fractions of enthalpy loss that occurs for each cascade.
        - `eta_ts`, which is the assumed total-to-static efficiency.
        - `eta_tt`, which is the assumed total-to-total efficiency.
        - `Ma_crit`, which is the assumed critical mash number.

    It can also be a dictionary containing the full set of initial guess that is provided directly to the solver. This
    require care as the user must have a complete knowledge of the different variables, and setup, of the initial guess that must be given that
    corresponds with the rest of the configuration file. If the initial guess is not given, it is set to a default value.
    For subsequent operation points, the function employs a strategy to use the closest previously computed operation point's solution
    as the initial guess. This approach is based on the heuristic that similar operation points have similar
    performance characteristics, which can improve convergence speed and robustness of the solution process.
    If the solution fails to converge, a set of initial guesses is provided to try other guesses (see `get_heuristic_guess_input`).

    The function returns a list of solver object for each operation point. This contain information on both solver related performance (see psv.NonlinearSystemSolver)
    and the object of the performance analysis problem (see CascadesNonlinearSystemProblem).

    Parameters
    ----------
    operation_points : list of dict or dict
        A list of operation points where each is a dictionary of parameters, or a dictionary of parameter
        ranges from which operation points will be generated.
    config : dict
        A dictionary containing necessary configuration options for computing performance at each operation point.
    initial_guess : optional
        A dictionary with the required elements to generate an initial guess (see description above).
    out_file : str, optional
        The name for the output Excel file. If not provided, a default name with a timestamp is generated.
    out_dir : str, optional
        The directory where the Excel file will be saved. Defaults to 'output'.
    stop_on_failure: bool, optional
        If true, the analysis stops if the solution fails to converge for an operating point.
    export_result : bool, optional
        If true, the result is exported to an excel file.

    Returns
    -------
    list
        A List of solver object for each operation point.

    """

    # Check if geometry is provided
    if config["geometry"] is None:
        raise ValueError("Geometry is not provided")

    # Check the type of operation_points argument
    if isinstance(operation_points, dict):
        # Convert ranges to a list of operation points
        operation_points = generate_operation_points(operation_points)
    elif not isinstance(operation_points, (list, np.ndarray)):
        msg = "operation_points must be either list of dicts or a dict with ranges."
        raise TypeError(msg)

    # Validate all operation points
    for operation_point in operation_points:
        validate_operation_point(operation_point)

    # Initialize lists to hold dataframes for each operation point
    operation_point_data = []
    overall_data = []
    plane_data = []
    cascade_data = []
    stage_data = []
    solver_data = []
    solution_data = []
    geometry_data = []
    solver_container = []

    # Loop through all operation points
    print_operation_points(operation_points)
    for i, operation_point in enumerate(operation_points):
        print()
        print(f" Computing operation point {i+1} of {len(operation_points)}")
        print_boundary_conditions(operation_point)

        try:
            # Define initial guess
            if i == 0:
                # Use default initial guess for the first operation point
                if initial_guess == None:
                    print("Using default initial guess")
                else:
                    print("Using user defined initial guess")
            else:
                closest_x, closest_index = find_closest_operation_point(
                    operation_point,
                    operation_points[:i],  # Use up to the previous point
                    solution_data[:i],  # Use solutions up to the previous point
                )
                print(f" Using solution from point {closest_index+1} as initial guess")
                initial_guess = closest_x

            # Compute performance
            solver, results = compute_single_operation_point(
                operation_point, initial_guess, config
            )

            # Retrieve solver data
            solver_status = {
                "completed": True,
                "success": solver.success,
                "message": solver.message,
                "grad_count": solver.convergence_history["grad_count"][-1],
                "func_count": solver.convergence_history["func_count"][-1],
                "func_count_total": solver.convergence_history["func_count_total"][-1],
                "norm_residual": solver.convergence_history["norm_residual"][-1],
                "norm_step": solver.convergence_history["norm_step"][-1],
            }

            # Collect results
            operation_point_data.append(pd.DataFrame([operation_point]))
            overall_data.append(results["overall"])
            plane_data.append(utils.flatten_dataframe(results["plane"]))
            cascade_data.append(utils.flatten_dataframe(results["cascade"]))
            stage_data.append(utils.flatten_dataframe(results["stage"]))
            geometry_data.append(utils.flatten_dataframe(results["geometry"]))
            solver_data.append(pd.DataFrame([solver_status]))
            solution_data.append(solver.problem.vars_real)
            solver_container.append(solver)

        except Exception as e:
            if stop_on_failure:
                raise Exception(e)
            else:
                print(f" Computation of point {i+1}/{len(operation_points)} failed")
                print(f" Error: {e}")

            # Retrieve solver data
            solver = None
            solver_status = {"completed": False}

            # Collect data
            operation_point_data.append(pd.DataFrame([operation_point]))
            overall_data.append(pd.DataFrame([{}]))
            plane_data.append(pd.DataFrame([{}]))
            cascade_data.append(pd.DataFrame([{}]))
            stage_data.append(pd.DataFrame([{}]))
            geometry_data.append(pd.DataFrame([{}]))
            solver_data.append(pd.DataFrame([solver_status]))
            solution_data.append([])
            solver_container.append(solver)

    # Dictionary to hold concatenated dataframes
    dfs = {
        "operation point": pd.concat(operation_point_data, ignore_index=True),
        "overall": pd.concat(overall_data, ignore_index=True),
        "plane": pd.concat(plane_data, ignore_index=True),
        "cascade": pd.concat(cascade_data, ignore_index=True),
        "stage": pd.concat(stage_data, ignore_index=True),
        "geometry": pd.concat(geometry_data, ignore_index=True),
        "solver": pd.concat(solver_data, ignore_index=True),
    }

    # Add 'operation_point' column to each dataframe
    for sheet_name, df in dfs.items():
        df.insert(0, "operation_point", range(1, 1 + len(df)))

    # Write dataframes to excel
    if export_results:
        # Create a directory to save simulation results
        if not os.path.exists(out_dir):
            os.makedirs(out_dir)

        # Define filename with unique date-time identifier
        if out_filename == None:
            current_time = datetime.datetime.now().strftime("%Y-%m-%d_%H-%M-%S")
            out_filename = f"performance_analysis_{current_time}"

        # Export simulation configuration as YAML file
        config_data = {k: v for k, v in config.items() if v}  # Filter empty entries
        config_data = utils.convert_numpy_to_python(config_data, precision=12)
        config_file = os.path.join(out_dir, f"{out_filename}.yaml")
        with open(config_file, "w") as file:
            yaml.dump(config_data, file, default_flow_style=False, sort_keys=False)

        # Export performance results in excel file
        filepath = os.path.join(out_dir, f"{out_filename}.xlsx")
        with pd.ExcelWriter(filepath, engine="openpyxl") as writer:
            for sheet_name, df in dfs.items():
                df.to_excel(writer, sheet_name=sheet_name, index=False)

        print(f" Performance data successfully written to {filepath}")

    # Print final report
    print_simulation_summary(solver_container)

    return solver_container


def compute_single_operation_point(
    operating_point,
    initial_guess,
    config,
):
    """
    Compute an operation point for a given set of boundary conditions using multiple solver methods and initial guesses.

    The initial guess make take three forms:

        - None, which generate a default initial guess
        - A dictionary which generates an initial guess through `compute_heuristic_initial_guess`. Required elements are:

            - `enthalpy_loss_fractions`, which is a list containing the assumed fractions of enthalpy loss that occurs for each cascade.
            - `eta_ts`, which is the assumed total-to-static efficiency.
            - `eta_tt`, which is the assumed total-to-total efficiency.
            - `Ma_crit`, which is the assumed critical mash number.

        - A dictionary containing the full set of variables needed to evaluate turbine performance. This option require that the user has complete knowledge of what are the required variables, and the setup of the inital guess dictionary for the given configuration.

    Parameters
    ----------
    boundary_conditions : dict
        A dictionary containing boundary conditions for the operation point.
    initial_guess : dict, optional
        A dictionary with the required elements to generate an initial guess (see description above).
    config : dict
        A dictionary containing necessary configuration options for computing performance at the operational point.

    Returns
    -------
    psv.NonlinearSystemSolver
        The solution object containing the results of the operation point calculation.

    """

    # Initialize problem object
    problem = CascadesNonlinearSystemProblem(config)
    # TODO: A limitation of defining a new problem for each operation point is that the geometry generated and checked once for every point
    # TODO: Performing the computations is not a big problem, but displaying the geometry report for every point can be very long.
    # TODO: Perhaps we could add options of verbosity and perhaps only display the full geometry report when it fails
    problem.update_boundary_conditions(operating_point)
    solver_options = copy.deepcopy(config["performance_analysis"]["solver_options"])

    initial_guesses = [initial_guess] + get_heuristic_guess_input(
        problem.geometry["number_of_cascades"]
    )
    methods_to_try = [solver_options["method"]] + [
        method for method in SOLVER_MAP.keys() if method != solver_options["method"]
    ]

    for initial_guess in initial_guesses:
        for method in methods_to_try:
            success = False
            x0 = problem.get_initial_guess(
                initial_guess
            )  # TODO: Roberto_17.05.2023: It seems we are not using this value. Why take it as output then?
            print(f" Trying to solve the problem using {SOLVER_MAP[method]} method")
            solver_options["method"] = method

            solver = psv.NonlinearSystemSolver(problem, **solver_options)
            # TODO: Roberto: add the option to use optimizers as solver depending on the method specified?
            # TODO: Roberto: at some point in the past we tried to solve the system of equations with SLSQP, right?

            try: 
                solver.solve(problem.x0)                
                
            except Exception as e:
                print(f" Error during solving: {e}")
                solver.success = False

            if solver.success:
                break
        if solver.success:
            break

    if not solver.success:
        print(" WARNING: All attempts failed to converge")
        # TODO: Add messages to Log file

    return solver, problem.results


def find_closest_operation_point(current_op_point, operation_points, solution_data):
    """
    Find the solution vector and index of the closest operation point in the historical data.

    Parameters
    ----------
    current_op_point : dict
        The current operation point we want to compare.
    operation_points : list of dict
        A list of historical operation points to search through.
    solution_data : list
        A list of solution vectors corresponding to each operation point.

    Returns
    -------
    tuple
        A tuple containing the closest solution vector and the one-based index of the closest operation point.

    """
    min_distance = float("inf")
    closest_point_x = None
    closest_index = None

    for i, op_point in enumerate(operation_points):
        distance = get_operation_point_distance(current_op_point, op_point)
        if distance < min_distance:
            min_distance = distance
            closest_point_x = solution_data[i]
            closest_index = i

    return closest_point_x, closest_index


def get_operation_point_distance(point_1, point_2, delta=1e-8):
    """
    Calculate the normalized distance between two operation points, with special consideration
    for angle measurements and prevention of division by zero for very small values.

    Parameters
    ----------
    point_1 : dict
        First operation point with numeric values.
    point_2 : dict
        Second operation point with numeric values.
    delta : float, optional
        A small constant to prevent division by zero. Default is 1e-8.

    Returns
    -------
    float
        The calculated normalized distance.
    """
    deviation_array = []
    for key in point_1:
        if isinstance(point_1[key], (int, float)) and key in point_2:
            value_1 = point_1[key]
            value_2 = point_2[key]

            if key == "alpha_in":
                # Handle angle measurements with absolute scale normalization
                deviation = np.abs(value_1 - value_2) / 90
            else:
                # Compute the relative difference with protection against division by zero
                max_val = max(abs(value_1), abs(value_2), delta)
                deviation = abs(value_1 - value_2) / max_val

            deviation_array.append(deviation)

    # Calculate the two-norm of the deviations
    return np.linalg.norm(deviation_array)


def generate_operation_points(performance_map):
    """
    Generates a list of dictionaries representing all possible combinations of
    operation points from a given performance map. The performance map is a
    dictionary where keys represent parameter names and values are the ranges
    of values for those parameters. The function ensures that the combinations
    are generated such that the parameters related to pressure ('p0_in' and
    'p_out') are the last ones to vary, effectively making them the first
    parameters to sweep through in the operation points.

    Parameters
    ----------
    - performance_map (dict): A dictionary with parameter names as keys and
      lists of parameter values as values.

    Returns
    -------
    - operation_points (list of dict): A list of dictionaries, each representing
      a unique combination of parameters from the performance_map.
    """

    # Make sure all values in the performance_map are iterables
    performance_map = {k: utils.ensure_iterable(v) for k, v in performance_map.items()}

    # Reorder performance map keys so first sweep is always through pressure
    priority_keys = ["p0_in", "p_out"]
    other_keys = [k for k in performance_map.keys() if k not in priority_keys]
    keys_order = other_keys + priority_keys
    performance_map = {
        k: performance_map[k] for k in keys_order if k in performance_map
    }

    # Create all combinations of operation points
    keys, values = zip(*performance_map.items())
    operation_points = [
        dict(zip(keys, combination)) for combination in itertools.product(*values)
    ]

    return operation_points


def validate_operation_point(op_point):
    """
    Validates that an operation point has exactly the required fields:
    'fluid_name', 'p0_in', 'T0_in', 'p_out', 'alpha_in', 'omega'.

    Parameters
    ----------
    op_point: dict
        A dictionary representing an operation point.

    Returns
    -------
    ValueError: If the dictionary does not contain the required fields or contains extra fields.
    """
    REQUIRED_FIELDS = {"fluid_name", "p0_in", "T0_in", "p_out", "alpha_in", "omega"}
    fields = set(op_point.keys())
    if fields != REQUIRED_FIELDS:
        missing = REQUIRED_FIELDS - fields
        extra = fields - REQUIRED_FIELDS
        raise ValueError(
            f"Operation point validation error: "
            f"Missing fields: {missing}, Extra fields: {extra}"
        )


# ------------------------------------------------------------------------------------------ #
# ------------------------------------------------------------------------------------------ #
# ------------------------------------------------------------------------------------------ #


class CascadesNonlinearSystemProblem(psv.NonlinearSystemProblem):
    """
    A class representing a nonlinear system problem for cascade analysis.

    This class is designed for solving nonlinear systems of equations related to cascade analysis.
    Derived classes must implement the `residual` method to evaluate the system of equations for a given set of decision variables.

    Additionally, specific problem classes can define the `get_jacobian` method to compute Jacobians.
    If this method is not present in the derived class, the solver will revert to using forward finite differences for Jacobian calculations.

    Attributes
    ----------
    fluid : FluidCoolProp_2Phase
        An instance of the FluidCoolProp_2Phase class representing the fluid properties.
    results : dict
        A dictionary to store results.
    boundary_conditions : dict
        A dictionary containing boundary condition data.
    geometry : dict
        A dictionary containing geometry-related data.
    model_options : dict
        A dictionary containing options related to the analysis model.
    reference_values : dict
        A dictionary containing reference values for calculations.
    vars_scaled
        A dicionary of scaled variables used to evaluate turbine performance.
    vars_real
        A dicionary of real variables used to evaluate turbine performance.

    Methods
    -------
    get_values(x)
        Evaluate the system of equations for a given set of decision variables.
    update_boundary_conditions(operation_point)
        Update the boundary conditions of the problem with the provided operation point.
    scale_values(variables, to_normalized = False)
        Convert values between normalized and real values.
    get_initial_guess(initial_guess = None)
        Determine the initial guess for the performance analysis based on the given parameters or default values.
    compute_heuristic_initial_guess(enthalpy_loss_fractions, eta_tt, eta_ts, Ma_crit)
        Compute the heuristic initial guess for the performance analysis based on the given parameters.
    print_simulation_summary(solvers)
        Print a formatted footer summarizing the performance of all operation points.
    print_boundary_conditions(BC)
        Print the boundary conditions.
    print_operation_points(operation_points)
        Prints a summary table of operation points scheduled for simulation.

    Examples
    --------
    Here's an example of how to derive from `CascadesNonlinearSystemProblem`::

        class MyCascadeProblem(CascadesNonlinearSystemProblem):
            def get_values(self, x):
                # Implement evaluation logic here
                pass
    """

    def __init__(self, config):
        """
        Initialize a CascadesNonlinearSystemProblem.

        Parameters
        ----------
        config : dict
            A dictionary containing case-specific data.
        """

        # Process turbine geometry
        geom.validate_turbine_geometry(config["geometry"])
        self.geometry = geom.calculate_full_geometry(config["geometry"])
        # self.geom_info = geom.check_turbine_geometry(self.geometry, display=True)

        # Initialize other attributes
        self.model_options = config["simulation_options"]
        self.keys = []

    def residual(self, x):
        """
        Evaluate the system of equations for a given set of decision variables.

        Parameters
        ----------
        x : array-like
            Vector of decision variables.

        Returns
        -------
        numpy nd.array
            An array containing residual values.
        """

        # Create dictionary of scaled variables
        self.vars_scaled = dict(zip(self.keys, x))

        # Create dictionary of real variables
        self.vars_real = self.scale_values(self.vars_scaled, to_normalized=False)

        # Evaluate cascade series
        self.results = flow.evaluate_axial_turbine(
            self.vars_scaled,
            self.boundary_conditions,
            self.geometry,
            self.fluid,
            self.model_options,
            self.reference_values,
        )

        return np.array(list(self.results["residuals"].values()))

    def update_boundary_conditions(self, operation_point):
        """
        Update the boundary conditions of the problem with the provided operation point.

        This method updates the boundary conditions attributes used to evaluate the turbine performance.
        It also initializes a Fluid object using the 'fluid_name' specified in the operation point.
        The method computes additional properties and reference values like stagnation properties at
        the inlet, exit static properties, spouting velocity, and reference mass flow rate.
        These are stored in the object's internal state for further use in calculations.

        Parameters
        ----------
        operation_point : dict
            A dictionary containing the boundary conditions defining the operation point. It must include the following keys:

            - `fluid_name` (str) : The name of the fluid to be used in the Fluid object.
            - `T0_in` (float): The inlet temperature (in Kelvin).
            - `p0_in` (float): The inlet pressure (in Pascals).
            - `p_out` (float): The outlet pressure (in Pascals).
            - `omega` (float): The rotational speed (in rad/s).
            - `alpha_in` (float): The inlet flow angle (in degrees).

        Returns
        -------
        None
            This method does not return a value but updates the internal state of the object.

        """

        # Define current operating point
        self.boundary_conditions = operation_point

        # Initialize fluid object
        self.fluid = props.Fluid(operation_point["fluid_name"], exceptions=True)

        # Rename variables
        p0_in = operation_point["p0_in"]
        T0_in = operation_point["T0_in"]
        p_out = operation_point["p_out"]

        # Compute stagnation properties at inlet
        state_in_stag = self.fluid.get_props(cp.PT_INPUTS, p0_in, T0_in)
        h0_in = state_in_stag["h"]
        s_in = state_in_stag["s"]

        # Store the inlet stagnation (h,s) for the first stage
        # TODO: Improve logic of implementation?
        self.boundary_conditions["h0_in"] = h0_in
        self.boundary_conditions["s_in"] = s_in

        # Calculate exit static properties for a isentropic expansion
        state_out_s = self.fluid.get_props(cp.PSmass_INPUTS, p_out, s_in)
        h_isentropic = state_out_s["h"]
        d_isentropic = state_out_s["d"]

        # Calculate exit static properties for a isenthalpic expansion
        state_out_h = self.fluid.get_props(cp.HmassP_INPUTS, h0_in, p_out)
        s_isenthalpic = state_out_h["s"]

        # Calculate spouting velocity
        v0 = np.sqrt(2 * (h0_in - h_isentropic))

        # Define a reference mass flow rate
        A_out = self.geometry["A_out"][-1]
        mass_flow_ref = A_out * v0 * d_isentropic

        # Define reference_values
        self.reference_values = {
            "s_range": s_isenthalpic - s_in,
            "s_min": s_in,
            "v0": v0,
            "h_out_s": h_isentropic,
            "d_out_s": d_isentropic,
            "mass_flow_ref": mass_flow_ref,
            "angle_range": 180,
            "angle_min": -90,
        }

        return

    def scale_values(self, variables, to_normalized=True):
        """
        Convert values between normalized and real values.

        Parameters
        ----------
        variables: dict
            A dictionary containing values to be scaled.
        to_real: bool
            If True, scale to real values; if False, scale to normalized values.

        Returns
        -------
        An array of values converted between scales.
        """

        # Load parameters
        v0 = self.reference_values["v0"]
        s_range = self.reference_values["s_range"]
        s_min = self.reference_values["s_min"]
        angle_range = self.reference_values["angle_range"]
        angle_min = self.reference_values["angle_min"]

        # Define dictionary of scaled values
        scaled_variables = {}

        for key, val in variables.items():
            if key.startswith("v") or key.startswith("w"):
                scaled_variables[key] = val / v0 if to_normalized else val * v0
            elif key.startswith("s"):
                scaled_variables[key] = (
                    (val - s_min) / s_range if to_normalized else val * s_range + s_min
                )
            elif key.startswith("b"):
                scaled_variables[key] = (
                    (val - angle_min) / angle_range
                    if to_normalized
                    else val * angle_range + angle_min
                )

        return scaled_variables

    def get_initial_guess(self, initial_guess=None):
        """
        Determine the initial guess for the performance analysis based on the given parameters or default values.

        The given `initial_guess` make take three forms:

            - None, which generate a default initial guess
            - A dictionary which generates an initial guess through `compute_heuristic_initial_guess`. Required elements are:

                - `enthalpy_loss_fractions`, which is a list containing the assumed fractions of enthalpy loss that occurs for each cascade.
                - `eta_ts`, which is the assumed total-to-static efficiency.
                - `eta_tt`, which is the assumed total-to-total efficiency.
                - `Ma_crit`, which is the assumed critical mash number.

            - A dictionary containing the full set of variables needed to evaluate turbine performance. This option require that the user has complete knowledge of what are the required variables, and the setup of the inital guess dictionary for the given configuration.

        The initial guess is scaled in this function.

        Parameters
        ----------
        initial_guess : dict, optional
            A dictionary containing the initial guess parameters.

        Returns
        -------
        dict
            Initial guess for the performance analysis.

        Raises
        ------
        ValueError
            If the provided initial_guess is invalid or incompatible with the model options.

        """

        number_of_cascades = self.geometry["number_of_cascades"]

        # Define initial guess
        if isinstance(initial_guess, dict):
            valid_keys_1 = ["enthalpy_loss_fractions", "eta_ts", "eta_tt", "Ma_crit"]
            valid_keys_2 = [
                "w_out",
                "s_out",
                "beta_out",
                "v*_in",
                "w*_throat",
                "s*_throat",
            ]
            valid_keys_2 = ["v_in"] + [
                f"{key}_{i+1}"
                for i in range(number_of_cascades)
                for key in valid_keys_2
            ]
            valid_keys_3 = [
                key for key in valid_keys_2 if not key.startswith("v*_in")
            ] + [f"beta*_throat_{i+1}" for i in range(number_of_cascades)]
            valid_keys_4 = [
                key
                for key in valid_keys_2
                if not (
                    key.startswith("v*_in")
                    or key.startswith("s*_throat")
                    or key.startswith("beta*_throat")
                )
            ]

            check = []
            check.append(set(valid_keys_1) == set(list(initial_guess.keys())))
            check.append(set(valid_keys_2) == set(list(initial_guess.keys())))
            check.append(set(valid_keys_3) == set(list(initial_guess.keys())))
            check.append(set(valid_keys_4) == set(list(initial_guess.keys())))

            if check[0]:
                enthalpy_loss_fractions = initial_guess["enthalpy_loss_fractions"]
                eta_tt = initial_guess["eta_tt"]
                eta_ts = initial_guess["eta_ts"]
                Ma_crit = initial_guess["Ma_crit"]

                # Check that eta_tt, eta_ts and Ma_crit is in reasonable range
                for label, variable in zip(
                    ["eta_tt", "eta_ts", "Ma_crit"], [eta_tt, eta_ts, Ma_crit]
                ):
                    if not 0 <= variable <= 1:
                        raise ValueError(f"{label} should be between {0} and {1}.")

                # Check if enthalpy_loss_fractions is a list or a NumPy array
                if not isinstance(enthalpy_loss_fractions, (list, np.ndarray)):
                    raise ValueError(
                        "enthalpy_loss_fractions must be a list or NumPy array"
                    )

                # Check that enthalpy_loss_fractions is of the same length as the number_of_cascades
                if len(enthalpy_loss_fractions) != number_of_cascades:
                    raise ValueError(
                        f"enthalpy_loss_fractions must be of length {number_of_cascades}"
                    )

                # Check that the sum of enthalpy loss fractions is 1
                sum_fractions = np.sum(enthalpy_loss_fractions)
                epsilon = 1e-8  # Small epsilon value for floating-point comparison
                if not np.isclose(sum_fractions, 1, atol=epsilon):
                    raise ValueError(
                        f"Sum of enthalpy_loss_fractions must be 1 (now: {sum_fractions})."
                    )

                print(" Generating heuristic initial guess with given parameters")
                initial_guess = self.compute_heuristic_initial_guess(
                    enthalpy_loss_fractions, eta_tt, eta_ts, Ma_crit
                )

            elif check[1]:
                # Check that set of input correspond with model option
                if (
                    not self.model_options["choking_model"]
                    == "evaluate_cascade_critical"
                ):
                    raise ValueError(
                        "Set of input corresponds with different choking_model (evaluate_cascade_critical)"
                    )

                # Check that all values are a number
                if not all(
                    isinstance(val, (int, float)) for val in initial_guess.values()
                ):
                    raise ValueError("All dictionary values must be a float or int")

            elif check[2]:
                # Check that set of input correspond with model option
                if not self.model_options["choking_model"] == "evaluate_cascade_throat":
                    raise ValueError(
                        "Set of input corresponds with different choking_model (evaluate_cascade_throat)"
                    )

                # Check that all values are a number
                if not all(
                    isinstance(val, (int, float)) for val in initial_guess.values()
                ):
                    raise ValueError("All dictionary values must be a float or int")

            elif check[3]:
                # Check that set of input correspond with model option
                if (
                    not self.model_options["choking_model"]
                    == "evaluate_cascade_isentropic_throat"
                ):
                    raise ValueError(
                        "Set of input corresponds with different choking_model (evaluate_cascade_isentropic_throat)"
                    )

                # Check that all values are a number
                if not all(
                    isinstance(val, (int, float)) for val in initial_guess.values()
                ):
                    raise ValueError("All dictionary values must be a float or int")

            else:
                raise ValueError(
                    f"Invalid keys provided for initial_guess. "
                    f"Valid keys include either:"
                    f"{valid_keys_1} \n"
                    f"{valid_keys_2} \n"
                    f"{valid_keys_3} \n"
                )

        elif initial_guess == None:
            enthalpy_loss_fractions = np.full(
                number_of_cascades, 1 / number_of_cascades
            )
            eta_tt = 0.9
            eta_ts = 0.8
            Ma_crit = 0.95

            # Compute initial guess using several approximations
            initial_guess = self.compute_heuristic_initial_guess(
                enthalpy_loss_fractions, eta_tt, eta_ts, Ma_crit
            )

        else:
            raise ValueError("Initial guess must be either None or a dictionary.")

        if self.model_options["choking_model"] == "evaluate_cascade_throat":
            initial_guess = {
                key: val
                for key, val in initial_guess.items()
                if not key.startswith("v*_in")
            }
        elif self.model_options["choking_model"] == "evaluate_cascade_critical":
            initial_guess = {
                key: val
                for key, val in initial_guess.items()
                if not key.startswith("beta*_throat")
            }
        elif (
            self.model_options["choking_model"] == "evaluate_cascade_isentropic_throat"
        ):
            initial_guess = {
                key: val
                for key, val in initial_guess.items()
                if not (
                    key.startswith("v*_in")
                    or key.startswith("s*_throat")
                    or key.startswith("beta*_throat")
                )
            }

        # Always normalize initial guess
        initial_guess_scaled = self.scale_values(initial_guess)

        # Store labels
        self.keys = initial_guess_scaled.keys()
        self.x0 = np.array(list(initial_guess_scaled.values()))

        return initial_guess_scaled

    def compute_heuristic_initial_guess(
        self, enthalpy_loss_fractions, eta_tt, eta_ts, Ma_crit
    ):
        """
        Compute the heuristic initial guess for the performance analysis based on the given parameters.

        This function calculates the heuristic initial guess based on the provided enthalpy loss fractions for each cascade,
        total-to-static and total-to-total efficiencies, and critical Mach number.

        Parameters
        ----------
        enthalpy_loss_fractions : array-like
            Enthalpy loss fractions for each cascade.
        eta_tt : float
            Total-to-total efficiency.
        eta_ts : float
            Total-to-static efficiency.
        Ma_crit : float
            Critical Mach number.

        Returns
        -------
        dict
            Heuristic initial guess for the performance analysis.

        """

        # Load object attributes
        geometry = self.geometry
        boundary_conditions = self.boundary_conditions
        fluid = self.fluid

        # Rename variables
        number_of_cascades = geometry["number_of_cascades"]
        p0_in = boundary_conditions["p0_in"]
        T0_in = boundary_conditions["T0_in"]
        alpha_in = boundary_conditions["alpha_in"]
        angular_speed = boundary_conditions["omega"]
        p_out = boundary_conditions["p_out"]
        h_out_s = self.reference_values["h_out_s"]

        # Calculate inlet stagnation state
        stagnation_properties_in = fluid.get_props(cp.PT_INPUTS, p0_in, T0_in)
        h0_in = stagnation_properties_in["h"]
        s_in = stagnation_properties_in["s"]
        rho0_in = stagnation_properties_in["d"]

        # Calculate exit enthalpy
        h0_out = h0_in - eta_ts * (h0_in - h_out_s)
        v_out = np.sqrt(2 * (h0_in - h_out_s - (h0_in - h0_out) / eta_tt))
        h_out = h0_out - 0.5 * v_out**2

        # Calculate exit static state for expansion with guessed efficiency
        static_properties_exit = fluid.get_props(cp.HmassP_INPUTS, h_out, p_out)
        s_out = static_properties_exit["s"]

        # Define entropy distribution
        entropy_distribution = np.linspace(s_in, s_out, number_of_cascades + 1)[1:]

        # Define enthalpy distribution
        total_enthalpy_loss = h0_in - h_out
        enthalpy_loss_per_cascade = [
            fraction * total_enthalpy_loss for fraction in enthalpy_loss_fractions
        ]
        enthalpy_distribution = [
            h0_in - sum(enthalpy_loss_per_cascade[: i + 1])
            for i in range(number_of_cascades)
        ]

        # Assums h0_in approx h_in for first inlet
        h_in = h0_in

        # Define initial guess dictionary
        initial_guess = {}

        for i in range(number_of_cascades):
            geometry_cascade = {
                key: values[i]
                for key, values in geometry.items()
                if key not in ["number_of_cascades", "number_of_stages"]
            }

            # Load enthalpy from initial guess
            h_out = enthalpy_distribution[i]

            # Load entropy from assumed entropy distribution
            s_out = entropy_distribution[i]

            # Rename necessary geometry
            theta_in = geometry_cascade["leading_edge_angle"]
            theta_out = geometry_cascade["gauging_angle"]
            A_out = geometry_cascade["A_out"]
            A_throat = geometry_cascade["A_throat"]
            A_in = geometry_cascade["A_in"]
            radius_mean_in = geometry_cascade["radius_mean_in"]
            radius_mean_throat = geometry_cascade["radius_mean_throat"]
            radius_mean_out = geometry_cascade["radius_mean_out"]

            # Calculate rothalpy at inlet of cascade
            blade_speed_in = angular_speed * (i % 2) * radius_mean_in
            if i == 0:
                h0_rel_in = h_in
                m_temp = rho0_in * A_in * math.cosd(alpha_in)
            else:
                v_in = np.sqrt(2 * (h0_in - h_in))
                velocity_triangle_in = flow.evaluate_velocity_triangle_in(
                    blade_speed_in, v_in, alpha_in
                )
                w_in = velocity_triangle_in["w"]
                h0_rel_in = h_in + 0.5 * w_in**2

            rothalpy = h0_rel_in - 0.5 * blade_speed_in**2

            # Calculate static state at cascade inlet
            static_state_in = fluid.get_props(cp.HmassSmass_INPUTS, h_in, s_in)
            rho_in = static_state_in["d"]

            # Calculate exit velocity from rothalpy and enthalpy distirbution
            blade_speed_out = angular_speed * (i % 2) * radius_mean_out
            h0_rel_out = rothalpy + 0.5 * blade_speed_out**2
            w_out = np.sqrt(2 * (h0_rel_out - h_out))
            velocity_triangle_out = flow.evaluate_velocity_triangle_out(
                blade_speed_out, w_out, theta_out
            )
            v_t_out = velocity_triangle_out["v_t"]
            v_m_out = velocity_triangle_out["v_m"]
            v_out = velocity_triangle_out["v"]
            h0_out = h_out + 0.5 * v_out**2

            # Calculate static state at cascade exit
            static_state_out = fluid.get_props(cp.HmassSmass_INPUTS, h_out, s_out)
            a_out = static_state_out["a"]
            rho_out = static_state_out["d"]

            # Calculate mass flow rate
            mass_flow = rho_out * v_m_out * A_out

            # Calculate throat velocity depending on subsonic or supersonic conditions
            if w_out < a_out * Ma_crit:
                w_throat = w_out
                s_throat = s_out
            else:
                w_throat = a_out * Ma_crit
                blade_speed_throat = angular_speed * (i % 2) * radius_mean_throat
                h0_rel_throat = rothalpy + 0.5 * blade_speed_throat**2
                h_throat = h0_rel_throat - 0.5 * w_throat**2
                rho_throat = rho_out
                static_state_throat = fluid.get_props(
                    cp.DmassHmass_INPUTS, rho_throat, h_throat
                )
                s_throat = static_state_throat["s"]

            # Calculate critical state
            w_throat_crit = a_out * Ma_crit
            h_throat_crit = (
                h0_rel_in - 0.5 * w_throat_crit**2
            )  # FIXME: h0_rel_in works less good?
            static_state_throat_crit = fluid.get_props(
                cp.HmassSmass_INPUTS, h_throat_crit, s_throat
            )
            rho_throat_crit = static_state_throat_crit["d"]
            m_crit = w_throat_crit * math.cosd(theta_out) * rho_throat_crit * A_throat
            w_m_in_crit = m_crit / rho_in / A_in
            w_in_crit = w_m_in_crit / math.cosd(
                theta_in
            )  # XXX Works better with metal angle than inlet flow angle?
            velocity_triangle_crit_in = flow.evaluate_velocity_triangle_out(
                blade_speed_in, w_in_crit, theta_in
            )
            v_in_crit = velocity_triangle_crit_in["v"]

            rho_out_crit = rho_throat_crit
            w_out_crit = m_crit / (rho_out_crit * A_out * math.cosd(theta_out))

            # Store initial guess
            index = f"_{i+1}"
            initial_guess.update(
                {
                    "w_out" + index: w_out,
                    "s_out" + index: s_out,
                    "beta_out"
                    + index: np.sign(theta_out) * math.arccosd(A_throat / A_out),
                    "v*_in" + index: v_in_crit,
                    "beta*_throat"
                    + index: np.sign(theta_out) * math.arccosd(A_throat / A_out),
                    "w*_throat" + index: w_throat_crit,
                    "s*_throat" + index: s_throat,
                }
            )

            # Update variables for next cascade
            if i != (number_of_cascades - 1):
                A_next = geometry["A_in"][i + 1]
                radius_mean_next = geometry["radius_mean_in"][i + 1]
                v_m_in = v_m_out * A_out / A_next
                v_t_in = v_t_out * radius_mean_out / radius_mean_next
                v_in = np.sqrt(v_m_in**2 + v_t_in**2)
                alpha_in = math.arctand(v_t_in / v_m_in)
                h0_in = h0_out
                h_in = h0_in - 0.5 * v_in**2
                s_in = s_out

        # Calculate inlet velocity from
        initial_guess["v_in"] = mass_flow / m_temp

        return initial_guess


def print_simulation_summary(solvers):
    """
    Print a formatted footer summarizing the performance of all operation points.

    This function processes a list of solver objects to provide a summary of the performance
    analysis calculations. It calculates and displays the number of successful points and a summary of
    simulation tme statistics. Additionally, it lists the indices of failed operation points, if any.

    The function is robust against solvers that failed and lack certain attributes like 'elapsed_time'.
    In such cases, these solvers are included in the count of failed operation points, but not in the
    calculation time statistics.

    Parameters
    ----------
    solvers : list
        A list of solver objects. Each solver object should contain attributes related to the
        calculation of an operation point, such as 'elapsed_time' and the 'solution' status.

    """

    # Initialize times list and track failed points
    times = []
    failed_points = []

    for i, solver in enumerate(solvers):
        # Check if the solver is not None and has the required attribute
        if solver and hasattr(solver, "elapsed_time"):
            times.append(solver.elapsed_time)
            if not solver.success:
                failed_points.append(i)
        else:
            # Handle failed solver or missing attributes
            failed_points.append(i)

    # Convert times to a numpy array for calculations
    times = np.asarray(times)
    total_points = len(solvers)

    # Define footer content
    width = 80
    separator = "-" * width
    lines_to_output = [
        "",
        separator,
        "Final summary of performance analysis calculations".center(width),
        separator,
        f" Simulation successful for {total_points - len(failed_points)} out of {total_points} points",
    ]

    # Add failed points message only if there are failed points
    if failed_points:
        lines_to_output.append(
            f"Failed operation points: {', '.join(map(str, failed_points))}"
        )

    # Add time statistics only if there are valid times
    if times.size > 0:
        lines_to_output.extend(
            [
                f" Average calculation time per operation point: {np.mean(times):.3f} seconds",
                f" Minimum calculation time of all operation points: {np.min(times):.3f} seconds",
                f" Maximum calculation time of all operation points: {np.max(times):.3f} seconds",
                f" Total calculation time for all operation points: {np.sum(times):.3f} seconds",
            ]
        )
    else:
        lines_to_output.append(" No valid calculation times available.")

    lines_to_output.append(separator)
    lines_to_output.append("")

    # Display to stdout
    for line in lines_to_output:
        print(line)


def print_boundary_conditions(BC):
    """
    Print the boundary conditions.

    This function prints the boundary conditions in a formatted manner. It takes a dictionary `BC`
    containing the following keys:

        - `fluid_name` (str): Name of the fluid.
        - `alpha_in` (float): Flow angle at inlet in degrees.
        - `T0_in` (float): Total temperature at inlet in Kelvin.
        - `p0_in` (float): Total pressure at inlet in Pascal.
        - `p_out` (float): Static pressure at outlet in Pascal.
        - `omega` (float): Angular speed in radians per second.

    Parameters
    ----------
    BC : dict
        A dictionary containing the boundary conditions.

    """

    column_width = 25  # Adjust this to your desired width
    print("-" * 80)
    print(" Operating point: ")
    print("-" * 80)
    print(f" {'Fluid: ':<{column_width}} {BC['fluid_name']:<}")
    print(f" {'Flow angle in: ':<{column_width}} {BC['alpha_in']:<.2f} deg")
    print(f" {'Total temperature in: ':<{column_width}} {BC['T0_in']-273.15:<.2f} degC")
    print(f" {'Total pressure in: ':<{column_width}} {BC['p0_in']/1e5:<.3f} bar")
    print(f" {'Static pressure out: ':<{column_width}} {BC['p_out']/1e5:<.3f} bar")
    print(f" {'Angular speed: ':<{column_width}} {BC['omega']*60/2/np.pi:<.1f} RPM")
    print("-" * 80)
    print()


def print_operation_points(operation_points):
    """
    Prints a summary table of operation points scheduled for simulation.

    This function takes a list of operation point dictionaries, formats them
    according to predefined specifications, applies unit conversions where
    necessary, and prints them in a neatly aligned table with headers and units.

    Parameters
    ----------
    - operation_points (list of dict): A list where each dictionary contains
      key-value pairs representing operation parameters and their corresponding
      values.

    Notes
    -----
    - This function assumes that all necessary keys exist within each operation
      point dictionary.
    - The function directly prints the output; it does not return any value.
    - Unit conversions are hardcoded and specific to known parameters.
    - If the units of the parameters change or if different parameters are added,
      the unit conversion logic and `field_specs` need to be updated accordingly.
    """
    length = 80
    index_width = 8
    output_lines = [
        "-" * length,
        " Summary of operation points scheduled for simulation",
        "-" * length,
    ]

    # Configuration for each field with specified width and decimal places
    field_specs = {
        "fluid_name": {"name": "Fluid", "unit": "", "width": 8},
        "alpha_in": {"name": "angle_in", "unit": "[deg]", "width": 10, "decimals": 1},
        "T0_in": {"name": "T0_in", "unit": "[degC]", "width": 12, "decimals": 2},
        "p0_in": {"name": "p0_in", "unit": "[kPa]", "width": 12, "decimals": 2},
        "p_out": {"name": "p_out", "unit": "[kPa]", "width": 12, "decimals": 2},
        "omega": {"name": "omega", "unit": "[RPM]", "width": 12, "decimals": 0},
    }

    # Create formatted header and unit strings using f-strings and field widths
    header_str = f"{'Index':>{index_width}}"  # Start with "Index" header
    unit_str = f"{'':>{index_width}}"  # Start with empty string for unit alignment

    for spec in field_specs.values():
        header_str += f" {spec['name']:>{spec['width']}}"
        unit_str += f" {spec['unit']:>{spec['width']}}"

    # Append formatted strings to the output lines
    output_lines.append(header_str)
    output_lines.append(unit_str)

    # Unit conversion functions
    def convert_units(key, value):
        if key == "T0_in":  # Convert Kelvin to Celsius
            return value - 273.15
        elif key == "omega":  # Convert rad/s to RPM
            return (value * 60) / (2 * np.pi)
        elif key == "alpha_in":  # Convert radians to degrees
            return np.degrees(value)
        elif key in ["p0_in", "p_out"]:  # Pa to kPa
            return value / 1e3
        return value

    # Process and format each operation point
    for index, op_point in enumerate(operation_points, start=1):
        row = [f"{index:>{index_width}}"]
        for key, spec in field_specs.items():
            value = convert_units(key, op_point[key])
            if isinstance(value, float):
                # Format floats with the specified width and number of decimal places
                row.append(f"{value:>{spec['width']}.{spec['decimals']}f}")
            else:
                # Format strings to the specified width without decimals
                row.append(f"{value:>{spec['width']}}")
        output_lines.append(" ".join(row))  # Ensure spaces between columns

    output_lines.append("-" * length)  # Add a closing separator line

    # Join the lines and print the output
    formatted_output = "\n".join(output_lines)

    for line in output_lines:
        print(line)
    return formatted_output



import numpy as np
from .. import math
from .. import utilities as utils


def validate_turbine_geometry(geom, display=False):
    """
    Performs validation of an axial turbine's geometry configuration.

    - Ensures all required geometry parameters are specified in 'geom'.
    - Verifies that no parameters beyond the required set are included.
    - Checks 'number_of_cascades' is an integer and correctly dictates sizes of other parameters.
    - Validates that all parameter arrays match the size of 'number_of_cascades'.
    - Asserts that all geometry parameter values are numeric (integers or floats).
    - Ensures non-angle parameters contain only non-negative values.
    - Angle parameters ('leading_edge_angle', 'stagger_angle') can be negative.

    This function is intended as a preliminary check before in-depth geometry analysis.

    Parameters
    ----------
    geom : dict
        The geometry configuration parameters for the turbine as a dictionary.

    Returns
    -------
    bool
        True if all validations pass, indicating a correctly structured geometry configuration.
    """

    # Define the list of keys that must be defined in the configuration file
    required_keys = {
        "cascade_type",
        "radius_hub_in",
        "radius_hub_out",
        "radius_tip_in",
        "radius_tip_out",
        "pitch",
        "chord",
        "stagger_angle",
        "opening",
        "leading_edge_diameter",
        "leading_edge_wedge_angle",
        "leading_edge_angle",
        "trailing_edge_thickness",
        "tip_clearance",
        "maximum_thickness",
        "throat_location_fraction",
    }

    # Angular variables need not be non-negative
    angle_keys = ["leading_edge_angle", "stagger_angle"]

    # Check that there are no missing or extra variables
    utils.validate_keys(geom, required_keys, required_keys)

    # Get the number of cascades
    number_of_cascades = len(geom["cascade_type"])

    # Check if cascade_type is correctly defined
    valid_types = np.array(["stator", "rotor"])
    invalid_types = geom["cascade_type"][~np.isin(geom["cascade_type"], valid_types)]
    if invalid_types.size > 0:
        invalid_types_str = ", ".join(invalid_types)
        raise ValueError(
            f"Invalid types in 'cascade_type': {invalid_types_str}. Only 'stator' and 'rotor' are allowed."
        )

    # Check array size and values
    for key, value in geom.items():
        # Skip the cascade_type in the loop
        if key in ["cascade_type", "geometry_type"]:
            continue

        # Check that all variables are numpy arrays
        if not isinstance(value, np.ndarray):
            raise TypeError(f"Parameter '{key}' must be a NumPy array.")

        else:
            if value.size != number_of_cascades:
                raise ValueError(
                    f"Size of '{key}' must be equal to number of cascades. Current value: {value}"
                )

        # Check that values of parameters are numeric
        if not math.all_numeric(value):
            raise ValueError(
                f"Parameter '{key}' must be an array of numeric types. Current value: {value}"
            )

        # Check that non-angle variables are not negative
        if key not in angle_keys and not math.all_non_negative(value):
            raise ValueError(
                f"Parameter '{key}' must be an array of non-negative numbers. Current value: {value}"
            )

    msg = "The structure of the turbine geometry parameters is valid"
    if display:
        print(msg)

    return msg


def calculate_throat_radius(radius_in, radius_out, throat_location_fraction):
    """
    Calculate the throat radius as a weighted average of the inlet and outlet radii.

    This approach to estimate the area at the throat is inspired by the method
    described in Eq. (3) of :cite:`ainley_method_1951`. The throat radius is computed
    as a linear combination of the inlet and outlet radii, weighted by the location of
    the throat expressed as a fraction of the axial length of the cascade.
    For instance, a `throat_location_fraction` of 5/6 would return the throat radius as:
    (1/6)*radius_in + (5/6)*radius out.

    Parameters
    ----------
    radius_in : np.array
        The radius values at the inlet sections.
    radius_out : np.array
        The radius values at the outlet sections.
    throat_location_fraction : float
        The fraction used to weight the inlet and outlet radii in the calculation.

    Returns
    -------
    np.array
        The calculated throat radius values, representing the weighted average of the
        inlet and outlet radii based on the specified throat location.
    """
    return (
        1 - throat_location_fraction
    ) * radius_in + throat_location_fraction * radius_out


def prepare_geometry(geometry, radius_type):
    """

    Convert the geometrical design variables to parameters suitable for turbine evaluation.

    For design optimization, the geometrical design variables are defined as ratios, e.g. hub-to-tip radius and aspect ratio.
    This function convert the set of geometrical design variables to a set which is suited for turbine evaluation.
    The function utilize `radius_type` to decide if the given radius is constant for hub, mean or tip throughout the turbine.

    Parameters
    ----------
    geometry : dict
        A dictionary with all necessary geometrical parameters.
    radius_type : str
        A string deciding if hub, mean or tip radius is constant throughout the turbine.

    Returns
    -------
    dict
        Set of geometry suited for turbine evaluation.

    """
    # Create a copy of the input dictionary to avoid mutating the original
    geom = geometry.copy()

    # Get number of cascades
    number_of_cascades = len(geom["cascade_type"])

    # Get number of stages
    if number_of_cascades > 1:
        if math.is_even(number_of_cascades):
            number_of_stages = int(number_of_cascades / 2)
        else:
            number_of_stages = int((number_of_cascades - 1) / 2)
    else:
        number_of_stages = 0

    # Extract initial
    radius = geom["radius"]
    hub_to_tip_in = geom["hub_tip_ratio_in"]
    hub_to_tip_out = geom["hub_tip_ratio_out"]

    # Compute radius at hub and tip
    if radius_type == "constant_mean":
        radius_mean = radius
        radius_tip_in = 2 * radius_mean / (1 + hub_to_tip_in)
        radius_tip_out = 2 * radius_mean / (1 + hub_to_tip_out)
        radius_hub_in = hub_to_tip_in * radius_tip_in
        radius_hub_out = hub_to_tip_out * radius_tip_out
        radius_mean_in = np.full_like(hub_to_tip_in, radius_mean)
        radius_mean_out = np.full_like(hub_to_tip_out, radius_mean)
    elif radius_type == "constant_hub":
        radius_hub = radius
        radius_tip_in = radius_hub / hub_to_tip_in
        radius_tip_out = radius_hub / hub_to_tip_out
        radius_mean_in = radius_tip_in * (1 + hub_to_tip_in) / 2
        radius_mean_out = radius_tip_out * (1 + hub_to_tip_out) / 2
        radius_hub_in = np.full_like(hub_to_tip_in, radius_hub)
        radius_hub_out = np.full_like(hub_to_tip_out, radius_hub)
    elif radius_type == "constant_tip":
        radius_tip = radius
        radius_hub_in = hub_to_tip_in * radius_tip
        radius_hub_out = hub_to_tip_out * radius_tip
        radius_tip_in = np.full_like(hub_to_tip_in, radius_tip)
        radius_tip_out = np.full_like(hub_to_tip_out, radius_tip)
        radius_mean_in = radius_tip_in * (1 + hub_to_tip_in) / 2
        radius_mean_out = radius_tip_out * (1 + hub_to_tip_out) / 2

    # Compute throat radius
    radius_mean_throat = calculate_throat_radius(
        radius_mean_in, radius_mean_out, geom["throat_location_fraction"]
    )
    radius_hub_throat = calculate_throat_radius(
        radius_hub_in, radius_hub_out, geom["throat_location_fraction"]
    )
    radius_tip_throat = calculate_throat_radius(
        radius_tip_in, radius_tip_out, geom["throat_location_fraction"]
    )

    # Compute blade heights
    height_in = radius_tip_in - radius_hub_in
    height_throat = radius_tip_throat - radius_hub_throat
    height_out = radius_tip_out - radius_hub_out
    height = (height_in + height_out) / 2

    # Compute chord
    chord = height / geom["aspect_ratio"]

    # Compute pitch
    pitch = geom["pitch_chord_ratio"] * chord

    # Compute maximum thickness
    gauging_angle = geom["gauging_angle"]
    blade_camber = abs(geom["leading_edge_angle"] - gauging_angle)
    maximum_thickness = np.array(
        [
            (
                0.15 * (blade_camber[i] <= 40)
                + (0.15 + 1.25e-3 * (blade_camber[i] - 40))
                * (40 < blade_camber[i] <= 120)
                + 0.25 * (blade_camber[i] > 120)
            )
            * chord[i]
            for i in range(len(blade_camber))
        ]
    )
    # Compute areas
    A_in = np.pi * (radius_tip_in**2 - radius_hub_in**2)
    A_out = np.pi * (radius_tip_out**2 - radius_hub_out**2)
    A_throat = A_out * math.cosd(gauging_angle)

    # Compute opening
    opening = A_throat * pitch / (2 * np.pi * radius_mean_throat * height_throat)

    # Compute trailing edge thickness
    trailing_edge_thickness = geom["trailing_edge_thickness_opening_ratio"] * opening

    # Compute stagger angle
    stagger_angle = 0.5 * (geom["leading_edge_angle"] + gauging_angle)

    # Define new geometry dictionary
    new_geometry = {
        "radius_hub_in": radius_hub_in,
        "radius_tip_in": radius_tip_in,
        "radius_hub_out": radius_hub_out,
        "radius_tip_out": radius_tip_out,
        "pitch": pitch,
        "chord": chord,
        "opening": opening,
        "maximum_thickness": maximum_thickness,
        "cascade_type": geom["cascade_type"],
        "leading_edge_angle": geom["leading_edge_angle"],
        "gauging_angle": gauging_angle,
        "stagger_angle": stagger_angle,
        "leading_edge_diameter": geom["leading_edge_diameter"],
        "leading_edge_wedge_angle": geom["leading_edge_wedge_angle"],
        "trailing_edge_thickness": trailing_edge_thickness,
        "tip_clearance": geom["tip_clearance"],
        "throat_location_fraction": geom["throat_location_fraction"],
    }

    return new_geometry


def calculate_full_geometry(geometry):
    """
    Computes the complete geometry of an axial turbine based on input geometric parameters.

    Parameters
    ----------
    geometry : dict
        A dictionary containing the input geometry parameters

    Returns
    -------
    dict
        A dictionary with both the original and newly computed geometry parameters.
    """

    # Create a copy of the input dictionary to avoid mutating the original
    geom = geometry.copy()

    # Get number of cascades
    number_of_cascades = len(geom["cascade_type"])

    # Get number of stages
    if number_of_cascades > 1:
        if math.is_even(number_of_cascades):
            number_of_stages = int(number_of_cascades / 2)
        else:
            number_of_stages = int((number_of_cascades - 1) / 2)
    else:
        number_of_stages = 0

    # Get the location of the throat as a fraction of meridional length
    throat_frac = geom["throat_location_fraction"]

    # Extract initial radii and compute throat values
    radius_hub_in = geom["radius_hub_in"]
    radius_hub_out = geom["radius_hub_out"]
    radius_tip_in = geom["radius_tip_in"]
    radius_tip_out = geom["radius_tip_out"]
    radius_hub_throat = calculate_throat_radius(
        radius_hub_in, radius_hub_out, throat_frac
    )
    radius_tip_throat = calculate_throat_radius(
        radius_tip_in, radius_tip_out, throat_frac
    )

    # Calculate shroud radius by adding clearance to tip radius
    radius_shroud_in = radius_tip_in + geom["tip_clearance"]
    radius_shroud_out = radius_tip_out + geom["tip_clearance"]
    radius_shroud_throat = calculate_throat_radius(
        radius_shroud_in, radius_shroud_out, throat_frac
    )

    # Calculate the inner and outer radii for the mean section
    radius_mean_in = (radius_tip_in + radius_hub_in) / 2
    radius_mean_out = (radius_tip_out + radius_hub_out) / 2
    radius_mean_throat = calculate_throat_radius(
        radius_mean_in, radius_mean_out, throat_frac
    )

    # Compute hub to tip radius ratio
    hub_tip_ratio_in = radius_hub_in / radius_tip_in
    hub_tip_ratio_out = radius_hub_out / radius_tip_out
    hub_tip_ratio_throat = radius_hub_throat / radius_tip_throat

    # Compute blade height at different sections
    height_in = radius_tip_in - radius_hub_in
    height_out = radius_tip_out - radius_hub_out
    height_throat = radius_tip_throat - radius_hub_throat
    height = (height_in + height_out) / 2

    # Compute areas for the full, in, out, and throat sections
    A_in = np.pi * (radius_tip_in**2 - radius_hub_in**2)
    A_out = np.pi * (radius_tip_out**2 - radius_hub_out**2)
    A_throat = (
        2 * np.pi * radius_mean_throat * height_throat * geom["opening"] / geom["pitch"]
    )

    # Gauging angle
    gauging_angle = math.arccosd(A_throat / A_out) * np.array(
        [(-1) ** i for i in range(number_of_cascades)]
    )

    # Compute axial chord
    axial_chord = geom["chord"] * math.cosd(geom["stagger_angle"])

    # Compute flaring angle
    flaring_angle = np.arctan((height_out - height_in) / axial_chord / 2)

    # Compute geometric ratios
    aspect_ratio = height / geom["chord"]
    pitch_chord_ratio = geom["pitch"] / geom["chord"]
    solidity = 1.0 / pitch_chord_ratio
    maximum_thickness_chord_ratio = geom["maximum_thickness"] / geom["chord"]
    trailing_edge_thickness_opening_ratio = (
        geom["trailing_edge_thickness"] / geom["opening"]
    )
    tip_clearance_height = geom["tip_clearance"] / height
    leading_edge_diameter_chord_ratio = geom["leading_edge_diameter"] / geom["chord"]

    # Create a dictionary with the newly computed parameters
    new_parameters = {
        "number_of_stages": number_of_stages,
        "number_of_cascades": number_of_cascades,
        "axial_chord": axial_chord,
        "radius_mean_in": radius_mean_in,
        "radius_mean_out": radius_mean_out,
        "radius_mean_throat": radius_mean_throat,
        "radius_hub_in": radius_hub_in,
        "radius_hub_out": radius_hub_out,
        "radius_hub_throat": radius_hub_throat,
        "radius_tip_in": radius_tip_in,
        "radius_tip_out": radius_tip_out,
        "radius_tip_throat": radius_tip_throat,
        "radius_shroud_in": radius_shroud_in,
        "radius_shroud_out": radius_shroud_out,
        "radius_shroud_throat": radius_shroud_throat,
        "hub_tip_ratio_in": hub_tip_ratio_in,
        "hub_tip_ratio_out": hub_tip_ratio_out,
        "hub_tip_ratio_throat": hub_tip_ratio_throat,
        "A_in": A_in,
        "A_out": A_out,
        "A_throat": A_throat,
        "height": height,
        "height_in": height_in,
        "height_throat": height_throat,
        "height_out": height_out,
        "flaring_angle": flaring_angle,
        "aspect_ratio": aspect_ratio,
        "pitch_chord_ratio": pitch_chord_ratio,
        "solidity": solidity,
        "maximum_thickness_chord_ratio": maximum_thickness_chord_ratio,
        "trailing_edge_thickness_opening_ratio": trailing_edge_thickness_opening_ratio,
        "tip_clearance_height_ratio": tip_clearance_height,
        "leading_edge_diameter_chord_ratio": leading_edge_diameter_chord_ratio,
        "gauging_angle": gauging_angle,
    }

    return {**geometry, **new_parameters}


def check_turbine_geometry(geometry, display=True):
    """
    Checks if the geometrical parameters are within the predefined recommended ranges.


    .. table:: Recommended Parameter Ranges and References

        +--------------------------+--------------+--------------+--------------------------------------------------+
        | Variable                 | Lower Limit  | Upper Limit  | Reference(s)                                     |
        +==========================+==============+==============+==================================================+
        | Blade chord              | 5.0 mm       | inf          | Manufacturing                                    |
        +--------------------------+--------------+--------------+--------------------------------------------------+
        | Blade height             | 5.0 mm       | inf          | Manufacturing                                    |
        +--------------------------+--------------+--------------+--------------------------------------------------+
        | Blade maximum thickness  | 1.0 mm       | inf          | Manufacturing                                    |
        +--------------------------+--------------+--------------+--------------------------------------------------+
        | Blade trailing edge      | 0.5 mm       | inf          | Manufacturing                                    |
        | thickness                |              |              |                                                  |
        +--------------------------+--------------+--------------+--------------------------------------------------+
        | Tip clearance            | 0.2 mm       | inf          | Manufacturing                                    |
        +--------------------------+--------------+--------------+--------------------------------------------------+
        | Hub to tip radius ratio  | 0.50         | 0.95         | Figure (6) of :cite:`kacker_mean_1982`           |
        +--------------------------+--------------+--------------+--------------------------------------------------+
        | Flaring angle            | -25 deg      | +25 deg      | Section (7) of :cite:`ainley_examination_1951`   |
        +--------------------------+--------------+--------------+--------------------------------------------------+
        | Aspect ratio             | 0.80         | 5.00         | Section (7.3) of :cite:`saravanamuttoo_gas_2008` |
        |                          |              |              | Figure (13) of :cite:`kacker_mean_1982`          |
        +--------------------------+--------------+--------------+--------------------------------------------------+
        | Pitch to chord ratio     | 0.30         | 1.10         | Figure (4) of :cite:`ainley_method_1951`         |
        +--------------------------+--------------+--------------+--------------------------------------------------+
        | Stagger angle            | -10 deg      | +70 deg      | Figure (5) of :cite:`kacker_mean_1982`           |
        +--------------------------+--------------+--------------+--------------------------------------------------+
        | Leading edge metal angle | -60 deg      | +25 deg      | Figure (5) of :cite:`kacker_mean_1982`           |
        +--------------------------+--------------+--------------+--------------------------------------------------+
        | Trailing edge metal angle| +40 deg      | +80 deg      | Figure (4) of :cite:`ainley_method_1951`         |
        +--------------------------+--------------+--------------+--------------------------------------------------+
        | Leading wedge angle      | +10 deg      | +60 deg      | Figure (2) from :cite:`benner_influence_1997`    |
        |                          |              |              | Figure (10) from :cite:`pritchard_eleven_1985`   |
        +--------------------------+--------------+--------------+--------------------------------------------------+
        | Leading edge diameter to | 0.03         | 0.30         | Table (1) :cite:`moustapha_improved_1990`        |
        | chord ratio              |              |              |                                                  |
        +--------------------------+--------------+--------------+--------------------------------------------------+
        | Maximum thickness to     | 0.05         | 0.30         | Figure (4) of :cite:`kacker_mean_1982`           |
        | chord ratio              |              |              |                                                  |
        +--------------------------+--------------+--------------+--------------------------------------------------+
        | Trailing edge thickness  | 0.00         | 0.40         | Figure (14) of :cite:`kacker_mean_1982`          |
        | to opening ratio         |              |              |                                                  |
        +--------------------------+--------------+--------------+--------------------------------------------------+
        | Tip clearance to height  | 0.00         | 0.05         | Figure (7) of :cite:`dunham_improvements_1970`   |
        | ratio                    |              |              |                                                  |
        +--------------------------+--------------+--------------+--------------------------------------------------+
        | Throat location fraction | 0.50         | 1.00         | Equation (3) of :cite:`ainley_method_1951`       |
        +--------------------------+--------------+--------------+--------------------------------------------------+

    The ranges of metal angles and stagger angles are reversed for rotor cascades

    Parameters
    ----------
    geometry : dict
        A dictionary containing the computed geometry of the turbine.

    Returns
    -------
    list
        A list of messages with a summary of the checks.
    """

    # Define good practice ranges for each parameter
    recommended_ranges = {
        "chord": {"min": 5e-3, "max": np.inf},
        "height": {"min": 5e-3, "max": np.inf},
        "maximum_thickness": {"min": 1e-3, "max": np.inf},
        "trailing_edge_thickness": {"min": 5e-4, "max": np.inf},
        "tip_clearance": {"min": 2e-4, "max": np.inf},
        "hub_tip_ratio": {"min": 0.50, "max": 0.95},
        "aspect_ratio": {"min": 0.8, "max": 5.0},
        "pitch_chord_ratio": {"min": 0.3, "max": 1.1},
        "stagger_angle": {"min": -10, "max": +70},
        "leading_edge_angle": {"min": -60, "max": +25},
        "leading_edge_wedge_angle": {"min": 10, "max": 60},
        "leading_edge_diameter_chord_ratio": {"min": 0.03, "max": 0.30},
        "maximum_thickness_chord_ratio": {"min": 0.05, "max": 0.30},
        "trailing_edge_thickness_opening_ratio": {"min": 0.00, "max": 0.40},
        "tip_clearance_height_ratio": {"min": 0.0, "max": 0.05},
        "throat_location_fraction": {"min": 0.5, "max": 1.0},
    }

    special_angles = ["leading_edge_angle", "stagger_angle"]

    # Get the the type of cascade
    cascade_types = geometry["cascade_type"]

    # Define report width
    report_width = 80

    # Define report title
    msgs = []
    msgs.append("-" * report_width)  # Horizontal line
    msgs.append("Axial turbine geometry report".center(report_width))
    msgs.append("-" * report_width)  # Horizontal line

    # Define table header with four columns
    table_header = f" {'Parameter':<34}{'Value':>8}{'Range':>20}{'In range?':>14}"
    msgs.append(table_header)
    msgs.append("-" * report_width)  # Horizontal line

    # Initialize a list to store variables outside the recommended range
    vars_outside_range = []

    # Iterate over each good practice parameter and perform checks
    for parameter, limits in recommended_ranges.items():
        if parameter == "cascade_type":
            continue  # Skip the cascade_type entry

        lb, ub = limits["min"], limits["max"]
        value_array = geometry[parameter]

        for index, value in enumerate(np.atleast_1d(value_array)):
            # Determine cascade type for special cases
            blade_type = cascade_types[index % len(cascade_types)]

            if parameter in special_angles and blade_type == "rotor":
                lb, ub = -ub, -lb  # Reverse limits for rotor blades

            if parameter == "tip_clearance":
                if blade_type == "stator":
                    lb = 0.0
                elif blade_type == "rotor":
                    lb = limits["min"]

            # Check if the value is within the recommended range
            in_range = lb <= value <= ub

            # Create a formatted message in table format using f-strings
            bounds = f"({lb:+0.4f}, {ub:+0.4f})"
            formatted_message = f" {parameter+'_'+str(index+1):<32}{value:>+8.4f}{bounds:>24}{str(in_range):>12}"
            msgs.append(formatted_message)

            # Append variables outside the recommended range to the list
            if not in_range:
                vars_outside_range.append(f"{parameter}_{index+1}")

    # Footer for the geometry report
    msgs.append("-" * report_width)  # Horizontal line

    if not vars_outside_range:
        msgs.append(
            " Geometry report summary: All parameters are within recommended ranges."
        )
    else:
        msgs.append(
            " Geometry report summary: Some parameters are outside recommended ranges."
        )
        for warning in vars_outside_range:
            msgs.append(f"     - {warning}")

    # Footer for the geometry report
    msgs.append("-" * report_width)  # Horizontal line

    # Concatenate lines into single string
    msg = "\n".join(msgs)

    # Display the report
    if display:
        print(msg)

    return msg


# TODO: Check that this function behaves as intended
# Make table of correct values and give reference to motivation
# TODO angles are correct for rotor/statos
# Tip clearance can be zero for rotor
# TODO angles in degree
# TODO flaring angle calculation
# TODO throat area calculation correct in cascade_series



from pydantic import BaseModel, ValidationError
from .turbo_configurations import AxialTurbine
from typing import Union
import yaml
import numbers
import numpy as np
from . import utilities

TURBOMACHINERIES = ["axial_turbine"]


def convert_configuration_options(config):
    """
    Processes configuration data by evaluating string expressions as numerical values and converting lists to numpy arrays.

    This function iteratively goes through the configuration dictionary and converts string representations of numbers
    (e.g., "1+2", "2*np.pi") into actual numerical values using Python's `eval` function. It also ensures that all numerical
    values are represented as Numpy types for consistency across the application.

    Parameters
    ----------
    config : dict
        The configuration data loaded from a YAML file, typically containing a mix of strings, numbers, and lists.

    Returns
    -------
    dict
        The postprocessed configuration data where string expressions are evaluated as numbers, and all numerical values
        are cast to corresponding NumPy types.

    Raises
    ------
    ConfigurationError
        If a list contains elements of different types after conversion, indicating an inconsistency in the expected data types.
    """

    def convert_strings_to_numbers(data):
        """
        Recursively converts string expressions within the configuration data to numerical values.

        This function handles each element of the configuration: dictionaries are traversed recursively, lists are processed
        element-wise, and strings are evaluated as numerical expressions. Non-string and valid numerical expressions are
        returned as is. The conversion supports basic mathematical operations and is capable of recognizing Numpy functions
        and constants when used in the strings.

        Parameters
        ----------
        data : dict, list, str, or number
            A piece of the configuration data that may contain strings representing numerical expressions.

        Returns
        -------
        The converted data, with all string expressions evaluated as numbers.
        """
        if isinstance(data, dict):
            return {
                key: convert_strings_to_numbers(value) for key, value in data.items()
            }
        elif isinstance(data, list):
            return [convert_strings_to_numbers(item) for item in data]
        elif isinstance(data, bool):
            return data
        elif isinstance(data, str):
            # Evaluate strings as numbers if possible
            try:
                data = eval(data)
                return convert_numbers_to_numpy(data)
            except (NameError, SyntaxError, TypeError):
                return data
        elif isinstance(data, numbers.Number):
            # Convert Python number types to corresponding NumPy number types
            return convert_numbers_to_numpy(data)
        else:
            return data

    def convert_numbers_to_numpy(data):
        """
        Casts Python native number types (int, float) to corresponding Numpy number types.

        This function ensures that all numeric values in the configuration are represented using Numpy types.
        It converts integers to `np.int64` and floats to `np.float64`.

        Parameters
        ----------
        data : int, float
            A numerical value that needs to be cast to a Numpy number type.

        Returns
        -------
        The same numerical value cast to the corresponding Numpy number type.

        """
        if isinstance(data, int):
            return np.int64(data)
        elif isinstance(data, float):
            return np.float64(data)
        else:
            return data

    def convert_to_arrays(data, parent_key=""):
        """
        Convert lists within the input data to Numpy arrays.

        Iterates through the input data recursively. If a list is encountered, the function checks if all elements are of the same type.
        If they are, the list is converted to a Numpy array. If the elements are of different types, a :obj:`ConfigurationError` is raised.

        Parameters
        ----------
        data : dict or list or any
            The input data which may contain lists that need to be converted to NumPy arrays. The data can be a dictionary (with recursive processing for each value), a list, or any other data type.
        parent_key : str, optional
            The key in the parent dictionary corresponding to `data`, used for error messaging. The default is an empty string.

        Returns
        -------
        dict or list or any
            The input data with lists converted to NumPy arrays. The type of return matches the type of `data`. Dictionaries and other types are returned unmodified.

        Raises
        ------
        ValueError
            If a list within `data` contains elements of different types. The error message includes the problematic list and the types of its elements.

        """
        if isinstance(data, dict):
            return {k: convert_to_arrays(v, parent_key=k) for k, v in data.items()}
        elif isinstance(data, list):
            if not data:  # Empty list
                return data
            first_type = type(data[0])
            if not all(isinstance(item, first_type) for item in data):
                element_types = [type(item) for item in data]
                raise ValueError(
                    f"Option '{parent_key}' contains elements of different types: {data}, "
                    f"types: {element_types}"
                )
            return np.array(data)
        else:
            return data

    # Convert the configuration options to Numpy arrays when relevant
    config = convert_strings_to_numbers(config)
    config = convert_to_arrays(config)

    return config


def object_to_dict(obj):
    """
    Recursively convert an object's attributes to a dictionary.

    This function takes an object and converts its attributes to a dictionary format.
    It handles nested dictionaries, lists, and objects with a `__dict__` attribute,
    recursively converting all attributes to dictionaries or arrays as appropriate.

    Parameters
    ----------
    obj : any
        The object to convert. This can be a dictionary, list, or an object
        with a `__dict__` attribute.

    Returns
    -------
    dict
        A dictionary representation of the object, where all nested attributes
        are recursively converted.
    """
    if isinstance(obj, dict):
        return {k: object_to_dict(v) for k, v in obj.items()}
    elif isinstance(obj, list):
        return np.array([object_to_dict(i) for i in obj])
    elif hasattr(obj, "__dict__"):
        return {k: object_to_dict(v) for k, v in vars(obj).items()}
    else:
        return obj


# Function to load and validate the configuration file
def load_config(config_file_path: str, print_summary=True):
    """
    Load and process a configuration file.

    This function reads a YAML configuration file, validates its contents,
    and converts it into a configuration object for turbomachinery analysis.
    The configuration can be printed as a summary if specified.

    Parameters
    ----------
    config_file_path : str
        Path to the YAML configuration file.
    print_summary : bool, optional
        Whether to print a summary of the loaded configuration (default is True).

    Returns
    -------
    dict
        A dictionary representation of the configuration object,
        with all nested attributes recursively converted.

    Raises
    ------
    FileNotFoundError
        If the configuration file is not found.
    ValueError
        If required fields are missing or contain invalid values.
    yaml.YAMLError
        If there is an error parsing the YAML file.

    """

    try:
        with open(config_file_path, "r") as config_file:
            # Load config file
            config_data = yaml.safe_load(config_file)

            # Convert string representation of numbers to number and lists to arrays
            config_data = convert_configuration_options(config_data)

            if "turbomachinery" not in config_data:
                raise ValueError("Missing 'turbomachinery' field in configuration.")

            turbomachinery_type = config_data["turbomachinery"]

            if turbomachinery_type == "axial_turbine":
                config = AxialTurbine(**config_data)
            # elif turbomachinery_type == 'radial_turbine':
            #     config =  RadialTurbineConfig(**config_data)
            else:
                raise ValueError(
                    f"Unknown turbomachinery type: {turbomachinery_type}. Available turbomachineries are: {TURBOMACHINERIES}"
                )

            # Convert configuration object to a nested dictionary
            config = object_to_dict(config)

            if config is not None:
                # Print configuration summary
                if print_summary:
                    succsess_message = "Configuration loaded successfully: "
                    dashed_line = "-" * len(succsess_message)
                    print(dashed_line)
                    print(succsess_message)
                    print(dashed_line)
                    utilities.print_dict(config)
                    print(dashed_line)
                    print("\n")

            return config

    except FileNotFoundError:
        print("Configuration file not found.")
    except ValidationError as e:
        print("Validation error in configuration file:")
        print(e)
    except yaml.YAMLError as e:
        print("Error parsing YAML file:")
        print(e)


def read_configuration_file(filename):
    """
    Reads a YAML configuration file and converts options to NumPy types when possible.

    Parameters
    ----------
    filename : str
        The path to the YAML configuration file.

    Returns
    -------
    dict
        A dictionary containing the configuration options, with values converted to NumPy types where applicable.

    Raises
    ------
    Exception
        If there is an error reading or parsing the configuration file.

    """
    # Read configuration file
    try:
        with open(filename, "r") as file:
            config = yaml.safe_load(file)
    except Exception as e:
        raise Exception(
            f"Error parsing configuration file: '{filename}'. Original error: {e}"
        )

    # Convert options to Numpy when possible
    config = convert_configuration_options(config)

    return config


# Example usage
if __name__ == "__main__":
    config_file_path = "config.yaml"  # Path to your YAML configuration file
    config = load_config(config_file_path, mode="optimization")
    if config:
        print("Configuration loaded successfully")



import os
import re
import yaml
import time
import logging
from datetime import datetime
import numpy as np


def convert_numpy_to_python(data, precision=10):
    """
    Recursively converts numpy arrays, scalars, and other numpy types to their Python counterparts
    and rounds numerical values to the specified precision.

    Parameters:
    - data: The numpy data to convert.
    - precision: The decimal precision to which float values should be rounded.

    Returns:
    - The converted data with all numpy types replaced by native Python types and float values rounded.
    """

    if data is None:
        return None

    if isinstance(data, dict):
        return {k: convert_numpy_to_python(v, precision) for k, v in data.items()}

    elif isinstance(data, list):
        return [convert_numpy_to_python(item, precision) for item in data]

    elif isinstance(data, np.ndarray):
        # If the numpy array has more than one element, it is iterable.
        if data.ndim > 0:
            return [convert_numpy_to_python(item, precision) for item in data.tolist()]
        else:
            # This handles the case of a numpy array with a single scalar value.
            return convert_numpy_to_python(data.item(), precision)

    elif isinstance(
        data,
        (np.integer, np.int_, np.intc, np.intp, np.int8, np.int16, np.int32, np.int64),
    ):
        return int(data.item())

    elif isinstance(data, (np.float_, np.float16, np.float32, np.float64)):
        return round(float(data.item()), precision)

    elif isinstance(data, np.bool_):
        return bool(data.item())

    elif isinstance(data, (np.str_, np.unicode_)):
        return str(data.item())

    # This will handle Python built-in types and other types that are not numpy.
    elif isinstance(data, (float, int, str, bool)):
        if isinstance(data, float):
            return round(data, precision)
        return data

    else:
        raise TypeError(f"Unsupported data type: {type(data)}")


def compare_contents_or_files(file_or_content_1, file_or_content_2):
    """
    Compare the content of two inputs, which can be either file paths or strings.

    This function accepts two arguments. Each argument can be:
    1. A file path pointing to a file containing text content.
    2. A string containing text content directly.

    If the argument is a file path that exists, the function reads its content.
    If the argument is a string, it's directly used for comparison.

    Parameters
    ----------
    file_or_content1 : str
        First input which can be a file path or string content.
    file_or_content2 : str
        Second input which can be a file path or string content.

    Returns
    -------
    bool
        True if the contents of the two inputs are identical, False otherwise.

    Examples
    --------
    >>> content_same("path/to/file1.txt", "path/to/file2.txt")
    True
    >>> content_same("Hello, world!", "path/to/file_with_hello_world_content.txt")
    True
    >>> content_same("Hello, world!", "Goodbye, world!")
    False
    """
    # If the first argument is a filepath and it exists, read its content
    if os.path.exists(file_or_content_1):
        with open(file_or_content_1, "r") as f1:
            file_or_content_1 = f1.read()

    # If the second argument is a filepath and it exists, read its content
    if os.path.exists(file_or_content_2):
        with open(file_or_content_2, "r") as f2:
            file_or_content_2 = f2.read()

    return file_or_content_1 == file_or_content_2


def create_logger(name, path=None, use_datetime=True):
    """
    Creates and configures a logging object for recording logs during program execution.

    Parameters
    ----------
    name : str
        Name of the log file. Allows for differentiation when analyzing logs from different components or runs of a program.
    path : str, optional
        Specifies the directory where the log files will be saved. By default, a directory named "logs"
        will be created in the current working directory (cwd).
    use_datetime : bool, optional
        Determines whether the log filename should have a unique datetime identifier appended. Default is True.

    Returns
    -------
    logger : object
        Configured logger object.

    Notes
    -----
    - By default, the function sets the log level to `INFO`, which means the logger will handle
      messages of level `INFO` and above (like `ERROR`, `WARNING`, etc.). The log entries will contain
      the timestamp, log level, and the actual log message.
    - When `use_datetime=True`, each log file will have a unique datetime identifier. This ensures
      traceability, avoids overwriting previous logs, allows chronological ordering of log files, and
      handles concurrency in multi-instance environments.
    """

    # Define logs directory if it is not provided
    if path is None:
        path = os.path.join(os.getcwd(), "logs")

    # Create logs directory if it does not exist
    if not os.path.exists(path):
        os.makedirs(path)

    # Define file name and path
    if use_datetime:
        current_time = datetime.now().strftime("%Y-%m-%d_%H-%M-%S")
        log_filename = os.path.join(path, f"{name}_{current_time}.log")
    else:
        log_filename = os.path.join(path, f"{name}.log")

    # Set logger configuration
    logging.basicConfig(
        filename=log_filename,
        level=logging.INFO,
        format="%(asctime)s - %(levelname)s - %(message)s",
    )

    # Create logger object
    logger = logging.getLogger()

    return logger


def find_latest_results_file(results_path, prefix="performance_analysis_"):
    """
    Retrieve all files that match the given prefix and extension .xlsx
    """
    files = sorted(
        [
            f
            for f in os.listdir(results_path)
            if f.startswith(prefix) and f.endswith(".xlsx")
        ]
    )

    # Return the first item from the sorted list, which is the latest file
    if files:
        return os.path.join(results_path, files[-1])
    else:
        raise FileNotFoundError(
            f"No Excel files found in directory '{results_path}' with prefix '{prefix}'"
        )


def wait_for_file(file_path, timeout=None, poll_interval=0.1):
    """
    Wait until the specified file is created.

    This function is used to wait until a file is created.

    Parameters
    ----------
    file_path : str
        Path to the file to wait for.
    timeout : float, optional
        Maximum time to wait in seconds. If None, waits indefinitely.
    poll_interval : int, optional
        Time interval between checks in seconds.

    Returns
    -------
    bool
        True if the file was found, False otherwise (only if timeout is set).
    """
    start_time = time.time()
    while True:
        if os.path.exists(file_path):
            return True
        if timeout is not None and time.time() - start_time > timeout:
            raise FileNotFoundError(f"Timeout waiting for file: {file_path}")
        time.sleep(poll_interval)


def add_string_to_keys(input_dict, suffix):
    """
    Add a suffix to each key in the input dictionary.

    Parameters
    ----------
        input_dict (dict): The input dictionary.
        suffix (str): The string to add to each key.

    Returns
    -------
        dict: A new dictionary with modified keys.

    Examples
    --------
        >>> input_dict = {'a': 1, 'b': 2, 'c': 3}
        >>> add_string_to_keys(input_dict, '_new')
        {'a_new': 1, 'b_new': 2, 'c_new': 3}
    """
    return {f"{key}{suffix}": value for key, value in input_dict.items()}


def check_for_unused_keys(dict_in, dict_name, raise_error=False):
    """
    Checks for unused parameters in the given dictionaries and sub-dictionaries.
    If unused items are found, prints them in a tree-like structure and optionally raises an error.

    Parameters
    ----------
    dict_in : dict
        The dictionary of parameters to check.
    dict_name : str
        The name of the dictionary variable.
    raise_error : bool, optional
        If True, raises an exception when unused items are found, otherwise just prints a warning.

    Returns
    -------
    None
    """
    if not is_dict_empty(dict_in):
        dict_str = print_dict(dict_in, indent=1, return_output=True)

        if raise_error:
            raise ValueError(
                f"Dictionary '{dict_name}' is not empty. Contains unused items:\n{dict_str}"
            )
        else:
            print(f"Warning: Dictionary '{dict_name}' contains unused items.")
            print(dict_str)


def is_dict_empty(data):
    """
    Recursively checks if a dictionary and all its sub-dictionaries are empty.

    Parameters
    ----------
    data : dict
        The dictionary to check.

    Returns
    -------
    bool
        True if the dictionary and all sub-dictionaries are empty, False otherwise.
    """
    if isinstance(data, dict):
        return all(is_dict_empty(v) for v in data.values()) if data else True
    return False  # Not a dictionary


def print_dict(data, indent=0, return_output=False):
    """
    Recursively prints nested dictionaries with indentation or returns the formatted string.

    Parameters
    ----------
    data : dict
        The dictionary to print.
    indent : int, optional
        The initial level of indentation for the keys of the dictionary, by default 0.
    return_output : bool, optional
        If True, returns the formatted string instead of printing it.

    Returns
    -------
    str or None
        The formatted string representation of the dictionary if return_output is True, otherwise None.
    """
    lines = []
    for key, value in data.items():
        line = "    " * indent + str(key) + ": "
        if isinstance(value, dict):
            if value:
                lines.append(line)
                lines.append(print_dict(value, indent + 1, True))
            else:
                lines.append(line + "{}")
        else:
            lines.append(line + str(value))

    output = "\n".join(lines)
    if return_output:
        return output
    else:
        print(output)


def print_object(obj):
    """
    Prints all attributes and methods of an object, sorted alphabetically.

    - Methods are identified as callable and printed with the prefix 'Method: '.
    - Attributes are identified as non-callable and printed with the prefix 'Attribute: '.

    Parameters
    ----------
    obj : object
        The object whose attributes and methods are to be printed.

    Returns
    -------
    None
        This function does not return any value. It prints the attributes and methods of the given object.

    """
    # Retrieve all attributes and methods
    attributes = dir(obj)

    # Sort them alphabetically, case-insensitive
    sorted_attributes = sorted(attributes, key=lambda x: x.lower())

    # Iterate over sorted attributes
    for attr in sorted_attributes:
        # Retrieve the attribute or method from the object
        attribute_or_method = getattr(obj, attr)

        # Check if it is callable (method) or not (attribute)
        if callable(attribute_or_method):
            print(f"Method: {attr}")
        else:
            print(f"Attribute: {attr}")


def validate_keys(checked_dict, required_keys, allowed_keys=None):
    """
    Validate the presence of required keys and check for any unexpected keys in a dictionary.

    Give required keys and allowed keys to have complete control
    Give required keys twice to check that the list of keys is necessary and sufficient
    Give only required keys to allow all extra additional key

    Parameters
    ----------
    checked_dict : dict
        The dictionary to be checked.
    required_keys : set
        A set of keys that are required in the dictionary.
    allowed_keys : set
        A set of keys that are allowed in the dictionary.

    Raises
    ------
    ConfigurationError
        If either required keys are missing or unexpected keys are found.
    """

    # Convert input lists to sets for set operations
    checked_keys = set(checked_dict.keys())
    required_keys = set(required_keys)

    # Set allowed_keys to all present keys if not provided
    if allowed_keys is None:
        allowed_keys = checked_keys
    else:
        allowed_keys = set(allowed_keys)

    # Check for extra and missing keys
    missing_keys = required_keys - checked_keys
    extra_keys = checked_keys - allowed_keys

    # Prepare error messages
    error_messages = []
    if missing_keys:
        error_messages.append(f"Missing required keys: {missing_keys}")
    if extra_keys:
        error_messages.append(f"Found unexpected keys: {extra_keys}")

    # Raise combined error if there are any issues
    if error_messages:
        raise DictionaryValidationError("; ".join(error_messages))


class DictionaryValidationError(Exception):
    """Exception raised for errors in the configuration options."""

    def __init__(self, message, key=None, value=None):
        self.message = message
        self.key = key
        self.value = value
        super().__init__(self._format_message())

    def _format_message(self):
        if self.key is not None and self.value is not None:
            return f"{self.message} Key: '{self.key}', Value: {self.value}"
        return self.message



import copy
from .. import utilities
from .. import fluid_properties as props

from .components import compression_process, expansion_process, heat_exchanger

COLORS_MATLAB = utilities.COLORS_MATLAB


def evaluate_cycle(
    variables,
    parameters,
    constraints,
    objective_function,
):
    # Create copies to not change the originals
    variables = copy.deepcopy(variables)
    parameters = copy.deepcopy(parameters)

    # Initialize fluid objects
    working_fluid = props.Fluid(
        **parameters.pop("working_fluid"), identifier="working_fluid"
    )
    heating_fluid = props.Fluid(
        **parameters.pop("heating_fluid"), identifier="heating_fluid"
    )
    cooling_fluid = props.Fluid(
        **parameters.pop("cooling_fluid"), identifier="cooling_fluid"
    )

    # Extract heat source/sink parameters and give short names
    T_source_out_max = parameters["heat_source"].pop("exit_temperature_max")
    T_source_out_min = parameters["heat_source"].pop("exit_temperature_min")
    T_sink_out_max = parameters["heat_sink"].pop("exit_temperature_max")
    T_sink_out_min = parameters["heat_sink"].pop("exit_temperature_min")
    p_source_out = parameters["heat_source"].pop("exit_pressure")
    p_sink_out = parameters["heat_sink"].pop("exit_pressure")

    # Compute coldest state at the heat source exit
    source_out_min = heating_fluid.set_state(
        props.PT_INPUTS, p_source_out, T_source_out_min
    )

    # Extract pressure drops and give short names
    dp_heater_h = parameters["heater"].pop("pressure_drop_hot_side")
    dp_heater_c = parameters["heater"].pop("pressure_drop_cold_side")
    dp_recup_h = parameters["recuperator"].pop("pressure_drop_hot_side")
    dp_recup_c = parameters["recuperator"].pop("pressure_drop_cold_side")
    dp_cooler_h = parameters["cooler"].pop("pressure_drop_hot_side")
    dp_cooler_c = parameters["cooler"].pop("pressure_drop_cold_side")

    # Extract design variables from dictionary (make sure all are used)
    turbine_inlet_p = variables.pop("turbine_inlet_pressure")
    turbine_inlet_h = variables.pop("turbine_inlet_enthalpy")
    compressor_inlet_p = variables.pop("compressor_inlet_pressure")
    compressor_inlet_h = variables.pop("compressor_inlet_enthalpy")
    recuperator_effectiveness = variables.pop("recuperator_effectiveness")
    heat_source_temperature_out = variables.pop("heat_source_exit_temperature")
    heat_sink_temperature_out = variables.pop("heat_sink_exit_temperature")

    # Evaluate  compressor
    dp = (1.0 - dp_heater_c) * (1.0 - dp_recup_c)
    compressor_outlet_p = turbine_inlet_p / dp
    compressor_eff = parameters["compressor"].pop("efficiency")
    compressor_eff_type = parameters["compressor"].pop("efficiency_type")
    compressor = compression_process(
        working_fluid,
        compressor_inlet_h,
        compressor_inlet_p,
        compressor_outlet_p,
        compressor_eff,
        compressor_eff_type,
    )

    # Evaluate turbine
    dp = (1.0 - dp_cooler_h) * (1.0 - dp_recup_h)
    turbine_outlet_p = compressor_inlet_p / dp
    turbine_efficiency = parameters["turbine"].pop("efficiency")
    turbine_efficiency_type = parameters["turbine"].pop("efficiency_type")
    turbine = expansion_process(
        working_fluid,
        turbine_inlet_h,
        turbine_inlet_p,
        turbine_outlet_p,
        turbine_efficiency,
        turbine_efficiency_type,
    )

    # Evaluate recuperator
    eps = recuperator_effectiveness
    T_in_hot = turbine["state_out"].T
    p_in_cold = compressor["state_out"].p
    h_in_cold = compressor["state_out"].h
    p_out_cold = p_in_cold * (1.0 - dp_recup_c)
    h_out_cold_ideal = working_fluid.set_state(props.PT_INPUTS, p_out_cold, T_in_hot).h
    h_out_cold_ideal = working_fluid.set_state(props.PT_INPUTS, p_out_cold, T_in_hot).h
    h_out_cold_actual = h_in_cold + eps * (h_out_cold_ideal - h_in_cold)
    p_in_hot = turbine["state_out"].p
    h_in_hot = turbine["state_out"].h
    p_out_hot = p_in_hot * (1.0 - dp_recup_h)
    h_out_hot = h_in_hot - (h_out_cold_actual - h_in_cold)
    num_elements = parameters["recuperator"].pop("num_elements")
    recuperator = heat_exchanger(
        working_fluid,
        h_in_hot,
        h_out_hot,
        p_in_hot,
        p_out_hot,
        working_fluid,
        h_in_cold,
        h_out_cold_actual,
        p_in_cold,
        p_out_cold,
        counter_current=True,
        num_steps=num_elements,
    )

    # Evaluate heater
    h_in_cold = recuperator["cold_side"]["state_out"].h
    p_in_cold = recuperator["cold_side"]["state_out"].p
    h_out_cold = turbine["state_in"].h
    p_out_cold = turbine["state_in"].p
    T_in_hot = parameters["heat_source"].pop("inlet_temperature")
    p_in_hot = parameters["heat_source"].pop("inlet_pressure")
    h_in_hot = heating_fluid.set_state(props.PT_INPUTS, p_in_hot, T_in_hot).h
    T_out_hot = heat_source_temperature_out
    p_out_hot = p_in_hot * (1 - dp_heater_h)
    h_out_hot = heating_fluid.set_state(props.PT_INPUTS, p_out_hot, T_out_hot).h
    num_elements = parameters["heater"].pop("num_elements")
    heater = heat_exchanger(
        heating_fluid,
        h_in_hot,
        h_out_hot,
        p_in_hot,
        p_out_hot,
        working_fluid,
        h_in_cold,
        h_out_cold,
        p_in_cold,
        p_out_cold,
        counter_current=True,
        num_steps=num_elements,
    )

    # Evaluate heat source pump
    h_in = heater["hot_side"]["state_out"].h
    p_in = heater["hot_side"]["state_out"].p
    p_out = p_source_out
    efficiency = parameters["heat_source_pump"].pop("efficiency")
    efficiency_type = parameters["heat_source_pump"].pop("efficiency_type")
    heat_source_pump = compression_process(
        heating_fluid,
        h_in,
        p_in,
        p_out,
        efficiency,
        efficiency_type,
    )

    # Evaluate heat sink pump
    T_in = parameters["heat_sink"].pop("inlet_temperature")
    p_in = parameters["heat_sink"].pop("inlet_pressure")
    h_in = cooling_fluid.set_state(props.PT_INPUTS, p_in, T_in).h
    p_out = p_sink_out / (1 - dp_cooler_c)
    efficiency = parameters["heat_sink_pump"].pop("efficiency")
    efficiency_type = parameters["heat_sink_pump"].pop("efficiency_type")
    heat_sink_pump = compression_process(
        cooling_fluid,
        h_in,
        p_in,
        p_out,
        efficiency,
        efficiency_type,
    )

    # Evaluate cooler
    p_in_cold = heat_sink_pump["state_out"].p
    h_in_cold = heat_sink_pump["state_out"].h
    p_out_cold = p_in_cold * (1 - dp_cooler_c)
    T_out_cold = heat_sink_temperature_out
    h_out_cold = cooling_fluid.set_state(props.PT_INPUTS, p_out_cold, T_out_cold).h
    h_in_hot = recuperator["hot_side"]["state_out"].h
    p_in_hot = recuperator["hot_side"]["state_out"].p
    h_out_hot = compressor["state_in"].h
    p_out_hot = compressor["state_in"].p
    num_elements = parameters["cooler"].pop("num_elements")
    cooler = heat_exchanger(
        working_fluid,
        h_in_hot,
        h_out_hot,
        p_in_hot,
        p_out_hot,
        cooling_fluid,
        h_in_cold,
        h_out_cold,
        p_in_cold,
        p_out_cold,
        counter_current=True,
        num_steps=num_elements,
    )

    # Compute mass flow rates
    W_net = parameters.pop("net_power")
    turbine_work = turbine["specific_work"]
    compression_work = compressor["specific_work"]
    m_total = W_net / (turbine_work - compression_work)
    m_source = m_total * heater["mass_flow_ratio"]
    m_sink = m_total / cooler["mass_flow_ratio"]

    # Add the mass flow to the components
    heater["hot_side"]["mass_flow"] = m_source
    heater["cold_side"]["mass_flow"] = m_total
    recuperator["hot_side"]["mass_flow"] = m_total
    recuperator["cold_side"]["mass_flow"] = m_total
    cooler["hot_side"]["mass_flow"] = m_total
    cooler["cold_side"]["mass_flow"] = m_sink
    turbine["mass_flow"] = m_total
    compressor["mass_flow"] = m_total
    heat_source_pump["mass_flow"] = m_source
    heat_sink_pump["mass_flow"] = m_sink

    # Summary of components
    components = {
        "turbine": turbine,
        "compressor": compressor,
        "recuperator": recuperator,
        "heater": heater,
        "cooler": cooler,
        "heat_source_pump": heat_source_pump,
        "heat_sink_pump": heat_sink_pump,
    }

    # Compute energy balances
    for name, component in components.items():
        if component["type"] == "heat_exchanger":
            hot = component["hot_side"]
            cold = component["cold_side"]
            Q1 = hot["mass_flow"] * (hot["state_in"].h - hot["state_out"].h)
            Q2 = cold["mass_flow"] * (cold["state_out"].h - cold["state_in"].h)
            component["heat_flow_rate"] = Q1
            component["heat_flow_rate_"] = Q2
            component["heat_balance"] = Q1 - Q2
            component["power"] = 0.0

        if component["type"] in ["compressor", "turbine"]:
            component["power"] = component["mass_flow"] * component["specific_work"]
            component["heat_flow_rate"] = 0.00

    # First-law analysis
    Q_in = heater["heat_flow_rate"]
    Q_out = cooler["heat_flow_rate"]
    W_out = turbine["power"]
    W_comp = compressor["power"]
    W_aux = heat_source_pump["power"] + heat_sink_pump["power"]
    W_in = W_comp + W_aux
    Q_in_max = m_source * (heater["hot_side"]["state_in"].h - source_out_min.h)
    cycle_efficiency = (W_out - W_in) / Q_in
    system_efficiency = (W_out - W_in) / Q_in_max
    backwork_ratio = W_comp / W_out
    energy_balance = (Q_in + W_comp) - (W_out + Q_out)  # Ignore pumps

    # Define dictionary with 1st Law analysis
    energy_analysis = {
        "heater_heat_flow": Q_in,
        "heater_heat_flow_max": Q_in_max,
        "recuperator_heat_flow": recuperator["heat_flow_rate"],
        "cooler_heat_flow": Q_out,
        "turbine_power": W_out,
        "compressor_power": W_comp,
        "heat_source_pump_power": heat_source_pump["power"],
        "heat_sink_pump_power": heat_sink_pump["power"],
        "net_cycle_power": W_net,
        "net_system_power": W_out - W_in,
        "mass_flow_heating_fluid": m_source,
        "mass_flow_working_fluid": m_total,
        "mass_flow_cooling_fluid": m_sink,
        "cycle_efficiency": cycle_efficiency,
        "system_efficiency": system_efficiency,
        "backwork_ratio": backwork_ratio,
        "energy_balance": energy_balance,
    }

    # Evaluate objective function and constraints
    output = {"components": components, "energy_analysis": energy_analysis}
    f = utilities.evaluate_objective_function(output, objective_function)
    c_eq, c_ineq = utilities.evaluate_constraints(output, constraints)
    # f = None
    # c_eq = None
    # c_ineq = None

    # Set colors for plotting
    heater["hot_side"]["color"] = COLORS_MATLAB[6]
    heater["cold_side"]["color"] = COLORS_MATLAB[1]
    recuperator["hot_side"]["color"] = COLORS_MATLAB[1]
    recuperator["cold_side"]["color"] = COLORS_MATLAB[1]
    cooler["hot_side"]["color"] = COLORS_MATLAB[1]
    cooler["cold_side"]["color"] = COLORS_MATLAB[0]
    turbine["color"] = COLORS_MATLAB[1]
    compressor["color"] = COLORS_MATLAB[1]

    # Check if any fixed parameter or design variable was not used
    utilities.check_for_unused_keys(parameters, "parameters", raise_error=True)
    utilities.check_for_unused_keys(variables, "variables", raise_error=True)

    # # Summary of components
    # components = {
    #     "turbine": turbine,
    #     "compressor": compressor,
    #     # "recuperator": recuperator,
    #     # "heater": heater,
    #     # "cooler": cooler,
    #     # "heat_source_pump": heat_source_pump,
    #     # "heat_sink_pump": heat_sink_pump,
    # }

    # Cycle performance summary
    output = {
        **output,
        "working_fluid": working_fluid,
        "heating_fluid": heating_fluid,
        "cooling_fluid": cooling_fluid,
        "components": components,
        "objective_function": f,
        "equality_constraints": c_eq,
        "inequality_constraints": c_ineq,
    }

    return output
